import sys

from colour.utilities.deprecation import ModuleAPI, build_API_changes
from colour.utilities.documentation import is_documentation_building

from colour.hints import Any

from .common import (
    Jab_to_JCh,
    JCh_to_Jab,
    COLOURSPACE_MODELS,
    COLOURSPACE_MODELS_AXIS_LABELS,
    COLOURSPACE_MODELS_DOMAIN_RANGE_SCALE_1_TO_REFERENCE,
)
from .cam02_ucs import (
    JMh_CIECAM02_to_CAM02LCD,
    CAM02LCD_to_JMh_CIECAM02,
    JMh_CIECAM02_to_CAM02SCD,
    CAM02SCD_to_JMh_CIECAM02,
    JMh_CIECAM02_to_CAM02UCS,
    CAM02UCS_to_JMh_CIECAM02,
    XYZ_to_CAM02LCD,
    CAM02LCD_to_XYZ,
    XYZ_to_CAM02SCD,
    CAM02SCD_to_XYZ,
    XYZ_to_CAM02UCS,
    CAM02UCS_to_XYZ,
)
from .cam16_ucs import (
    JMh_CAM16_to_CAM16LCD,
    CAM16LCD_to_JMh_CAM16,
    JMh_CAM16_to_CAM16SCD,
    CAM16SCD_to_JMh_CAM16,
    JMh_CAM16_to_CAM16UCS,
    CAM16UCS_to_JMh_CAM16,
    XYZ_to_CAM16LCD,
    CAM16LCD_to_XYZ,
    XYZ_to_CAM16SCD,
    CAM16SCD_to_XYZ,
    XYZ_to_CAM16UCS,
    CAM16UCS_to_XYZ,
)
from .cie_xyy import (
    XYZ_to_xyY,
    xyY_to_XYZ,
    xy_to_xyY,
    xyY_to_xy,
    xy_to_XYZ,
    XYZ_to_xy,
)
from .cie_lab import XYZ_to_Lab, Lab_to_XYZ, Lab_to_LCHab, LCHab_to_Lab
from .cie_luv import (
    XYZ_to_Luv,
    Luv_to_XYZ,
    Luv_to_uv,
    uv_to_Luv,
    Luv_uv_to_xy,
    xy_to_Luv_uv,
    Luv_to_LCHuv,
    LCHuv_to_Luv,
)
from .cie_ucs import (
    XYZ_to_UCS,
    UCS_to_XYZ,
    UCS_to_uv,
    uv_to_UCS,
    UCS_uv_to_xy,
    xy_to_UCS_uv,
)
from .cie_uvw import XYZ_to_UVW, UVW_to_XYZ
from .din99 import Lab_to_DIN99, DIN99_to_Lab, XYZ_to_DIN99, DIN99_to_XYZ
from .hdr_cie_lab import (
    HDR_CIELAB_METHODS,
    XYZ_to_hdr_CIELab,
    hdr_CIELab_to_XYZ,
)
from .hunter_lab import (
    XYZ_to_K_ab_HunterLab1966,
    XYZ_to_Hunter_Lab,
    Hunter_Lab_to_XYZ,
)
from .hunter_rdab import XYZ_to_Hunter_Rdab, Hunter_Rdab_to_XYZ
from .icacb import XYZ_to_ICaCb, ICaCb_to_XYZ
from .igpgtg import XYZ_to_IgPgTg, IgPgTg_to_XYZ
from .ipt import XYZ_to_IPT, IPT_to_XYZ, IPT_hue_angle
from .jzazbz import (
    IZAZBZ_METHODS,
    XYZ_to_Izazbz,
    XYZ_to_Jzazbz,
    Izazbz_to_XYZ,
    Jzazbz_to_XYZ,
)
from .hdr_ipt import HDR_IPT_METHODS, XYZ_to_hdr_IPT, hdr_IPT_to_XYZ
from .oklab import XYZ_to_Oklab, Oklab_to_XYZ
from .osa_ucs import XYZ_to_OSA_UCS, OSA_UCS_to_XYZ
from .prolab import XYZ_to_ProLab, ProLab_to_XYZ
from .datasets import (
    DATA_MACADAM_1942_ELLIPSES,
    CCS_ILLUMINANT_POINTER_GAMUT,
    DATA_POINTER_GAMUT_VOLUME,
    CCS_POINTER_GAMUT_BOUNDARY,
)
from .rgb import (
    normalised_primary_matrix,
    chromatically_adapted_primaries,
    primaries_whitepoint,
    RGB_luminance_equation,
    RGB_luminance,
)
from .rgb import RGB_Colourspace
from .rgb import XYZ_to_RGB, RGB_to_XYZ
from .rgb import matrix_RGB_to_RGB, RGB_to_RGB
from .rgb import (
    CV_range,
    legal_to_full,
    full_to_legal,
    gamma_function,
    log_encoding_ACESproxy,
    log_decoding_ACESproxy,
    log_encoding_ACEScc,
    log_decoding_ACEScc,
    log_encoding_ACEScct,
    log_decoding_ACEScct,
    oetf_ARIBSTDB67,
    oetf_inverse_ARIBSTDB67,
    log_encoding_ALEXALogC,
    log_decoding_ALEXALogC,
    oetf_BlackmagicFilmGeneration5,
    oetf_inverse_BlackmagicFilmGeneration5,
    log_encoding_CanonLog,
    log_decoding_CanonLog,
    log_encoding_CanonLog2,
    log_decoding_CanonLog2,
    log_encoding_CanonLog3,
    log_decoding_CanonLog3,
    log_encoding_Cineon,
    log_decoding_Cineon,
    oetf_DaVinciIntermediate,
    oetf_inverse_DaVinciIntermediate,
    eotf_inverse_DCDM,
    eotf_DCDM,
    eotf_inverse_DICOMGSDF,
    eotf_DICOMGSDF,
    log_encoding_DJIDLog,
    log_decoding_DJIDLog,
    exponent_function_basic,
    exponent_function_monitor_curve,
    log_encoding_FilmicPro6,
    log_decoding_FilmicPro6,
    log_encoding_FilmLightTLog,
    log_decoding_FilmLightTLog,
    log_encoding_Protune,
    log_decoding_Protune,
    oetf_BT601,
    oetf_inverse_BT601,
    oetf_BT709,
    oetf_inverse_BT709,
    eotf_inverse_BT1886,
    eotf_BT1886,
    eotf_inverse_BT2020,
    eotf_BT2020,
    eotf_inverse_ST2084,
    eotf_ST2084,
    oetf_PQ_BT2100,
    oetf_inverse_PQ_BT2100,
    eotf_PQ_BT2100,
    eotf_inverse_PQ_BT2100,
    ootf_PQ_BT2100,
    ootf_inverse_PQ_BT2100,
    oetf_HLG_BT2100,
    oetf_inverse_HLG_BT2100,
    BT2100_HLG_EOTF_METHODS,
    eotf_HLG_BT2100,
    BT2100_HLG_EOTF_INVERSE_METHODS,
    eotf_inverse_HLG_BT2100,
    BT2100_HLG_OOTF_METHODS,
    ootf_HLG_BT2100,
    BT2100_HLG_OOTF_INVERSE_METHODS,
    ootf_inverse_HLG_BT2100,
    linear_function,
    logarithmic_function_basic,
    logarithmic_function_quasilog,
    logarithmic_function_camera,
    log_encoding_Log2,
    log_decoding_Log2,
    log_encoding_Panalog,
    log_decoding_Panalog,
    log_encoding_VLog,
    log_decoding_VLog,
    log_encoding_FLog,
    log_decoding_FLog,
    log_encoding_NLog,
    log_decoding_NLog,
    log_encoding_PivotedLog,
    log_decoding_PivotedLog,
    log_encoding_REDLog,
    log_decoding_REDLog,
    log_encoding_REDLogFilm,
    log_decoding_REDLogFilm,
    LOG3G10_ENCODING_METHODS,
    LOG3G10_DECODING_METHODS,
    log_encoding_Log3G10,
    log_decoding_Log3G10,
    log_encoding_Log3G12,
    log_decoding_Log3G12,
    cctf_encoding_ROMMRGB,
    cctf_decoding_ROMMRGB,
    cctf_encoding_ProPhotoRGB,
    cctf_decoding_ProPhotoRGB,
    cctf_encoding_RIMMRGB,
    cctf_decoding_RIMMRGB,
    log_encoding_ERIMMRGB,
    log_decoding_ERIMMRGB,
    oetf_SMPTE240M,
    eotf_SMPTE240M,
    log_encoding_SLog,
    log_decoding_SLog,
    log_encoding_SLog2,
    log_decoding_SLog2,
    log_encoding_SLog3,
    log_decoding_SLog3,
    eotf_inverse_sRGB,
    eotf_sRGB,
    log_encoding_ViperLog,
    log_decoding_ViperLog,
)
from .rgb import (
    LOG_ENCODINGS,
    log_encoding,
    LOG_DECODINGS,
    log_decoding,
    OETFS,
    oetf,
    OETF_INVERSES,
    oetf_inverse,
    EOTFS,
    eotf,
    EOTF_INVERSES,
    eotf_inverse,
    CCTF_ENCODINGS,
    cctf_encoding,
    CCTF_DECODINGS,
    cctf_decoding,
    OOTFS,
    ootf,
    OOTF_INVERSES,
    ootf_inverse,
)
from .rgb import (
    RGB_COLOURSPACES,
    RGB_COLOURSPACE_ACES2065_1,
    RGB_COLOURSPACE_ACESCC,
    RGB_COLOURSPACE_ACESCCT,
    RGB_COLOURSPACE_ACESPROXY,
    RGB_COLOURSPACE_ACESCG,
    RGB_COLOURSPACE_ADOBE_RGB1998,
    RGB_COLOURSPACE_ADOBE_WIDE_GAMUT_RGB,
    RGB_COLOURSPACE_ALEXA_WIDE_GAMUT,
    RGB_COLOURSPACE_APPLE_RGB,
    RGB_COLOURSPACE_BEST_RGB,
    RGB_COLOURSPACE_BETA_RGB,
    RGB_COLOURSPACE_BLACKMAGIC_WIDE_GAMUT,
    RGB_COLOURSPACE_BT470_525,
    RGB_COLOURSPACE_BT470_625,
    RGB_COLOURSPACE_BT709,
    RGB_COLOURSPACE_BT2020,
    RGB_COLOURSPACE_CIE_RGB,
    RGB_COLOURSPACE_CINEMA_GAMUT,
    RGB_COLOURSPACE_COLOR_MATCH_RGB,
    RGB_COLOURSPACE_DAVINCI_WIDE_GAMUT,
    RGB_COLOURSPACE_DCDM_XYZ,
    RGB_COLOURSPACE_DCI_P3,
    RGB_COLOURSPACE_DCI_P3_P,
    RGB_COLOURSPACE_DISPLAY_P3,
    RGB_COLOURSPACE_DJI_D_GAMUT,
    RGB_COLOURSPACE_DON_RGB_4,
    RGB_COLOURSPACE_ECI_RGB_V2,
    RGB_COLOURSPACE_EKTA_SPACE_PS_5,
    RGB_COLOURSPACE_FILMLIGHT_E_GAMUT,
    RGB_COLOURSPACE_PROTUNE_NATIVE,
    RGB_COLOURSPACE_MAX_RGB,
    RGB_COLOURSPACE_N_GAMUT,
    RGB_COLOURSPACE_P3_D65,
    RGB_COLOURSPACE_PAL_SECAM,
    RGB_COLOURSPACE_RED_COLOR,
    RGB_COLOURSPACE_RED_COLOR_2,
    RGB_COLOURSPACE_RED_COLOR_3,
    RGB_COLOURSPACE_RED_COLOR_4,
    RGB_COLOURSPACE_DRAGON_COLOR,
    RGB_COLOURSPACE_DRAGON_COLOR_2,
    RGB_COLOURSPACE_RED_WIDE_GAMUT_RGB,
    RGB_COLOURSPACE_ROMM_RGB,
    RGB_COLOURSPACE_RIMM_RGB,
    RGB_COLOURSPACE_ERIMM_RGB,
    RGB_COLOURSPACE_PROPHOTO_RGB,
    RGB_COLOURSPACE_RUSSELL_RGB,
    RGB_COLOURSPACE_SHARP_RGB,
    RGB_COLOURSPACE_SMPTE_240M,
    RGB_COLOURSPACE_SMPTE_C,
    RGB_COLOURSPACE_NTSC1953,
    RGB_COLOURSPACE_NTSC1987,
    RGB_COLOURSPACE_S_GAMUT,
    RGB_COLOURSPACE_S_GAMUT3,
    RGB_COLOURSPACE_S_GAMUT3_CINE,
    RGB_COLOURSPACE_VENICE_S_GAMUT3,
    RGB_COLOURSPACE_VENICE_S_GAMUT3_CINE,
    RGB_COLOURSPACE_sRGB,
    RGB_COLOURSPACE_V_GAMUT,
    RGB_COLOURSPACE_XTREME_RGB,
    RGB_COLOURSPACE_F_GAMUT,
)

from .rgb import XYZ_to_sRGB, sRGB_to_XYZ
from .rgb import (
    RGB_to_HSV,
    HSV_to_RGB,
    RGB_to_HSL,
    HSL_to_RGB,
    RGB_to_HCL,
    HCL_to_RGB,
)
from .rgb import RGB_to_CMY, CMY_to_RGB, CMY_to_CMYK, CMYK_to_CMY
from .rgb import RGB_to_IHLS, IHLS_to_RGB
from .rgb import RGB_to_Prismatic, Prismatic_to_RGB
from .rgb import (
    WEIGHTS_YCBCR,
    matrix_YCbCr,
    offset_YCbCr,
    RGB_to_YCbCr,
    YCbCr_to_RGB,
    RGB_to_YcCbcCrc,
    YcCbcCrc_to_RGB,
)
from .rgb import RGB_to_YCoCg, YCoCg_to_RGB
from .rgb import RGB_to_ICtCp, ICtCp_to_RGB, XYZ_to_ICtCp, ICtCp_to_XYZ

__all__ = [
    "Jab_to_JCh",
    "JCh_to_Jab",
    "COLOURSPACE_MODELS",
    "COLOURSPACE_MODELS_AXIS_LABELS",
    "COLOURSPACE_MODELS_DOMAIN_RANGE_SCALE_1_TO_REFERENCE",
]
__all__ += [
    "JMh_CIECAM02_to_CAM02LCD",
    "CAM02LCD_to_JMh_CIECAM02",
    "JMh_CIECAM02_to_CAM02SCD",
    "CAM02SCD_to_JMh_CIECAM02",
    "JMh_CIECAM02_to_CAM02UCS",
    "CAM02UCS_to_JMh_CIECAM02",
    "XYZ_to_CAM02LCD",
    "CAM02LCD_to_XYZ",
    "XYZ_to_CAM02SCD",
    "CAM02SCD_to_XYZ",
    "XYZ_to_CAM02UCS",
    "CAM02UCS_to_XYZ",
]
__all__ += [
    "JMh_CAM16_to_CAM16LCD",
    "CAM16LCD_to_JMh_CAM16",
    "JMh_CAM16_to_CAM16SCD",
    "CAM16SCD_to_JMh_CAM16",
    "JMh_CAM16_to_CAM16UCS",
    "CAM16UCS_to_JMh_CAM16",
    "XYZ_to_CAM16LCD",
    "CAM16LCD_to_XYZ",
    "XYZ_to_CAM16SCD",
    "CAM16SCD_to_XYZ",
    "XYZ_to_CAM16UCS",
    "CAM16UCS_to_XYZ",
]
__all__ += [
    "XYZ_to_xyY",
    "xyY_to_XYZ",
    "xy_to_xyY",
    "xyY_to_xy",
    "xy_to_XYZ",
    "XYZ_to_xy",
]
__all__ += [
    "XYZ_to_Lab",
    "Lab_to_XYZ",
    "Lab_to_LCHab",
    "LCHab_to_Lab",
]
__all__ += [
    "XYZ_to_Luv",
    "Luv_to_XYZ",
    "Luv_to_uv",
    "uv_to_Luv",
    "Luv_uv_to_xy",
    "xy_to_Luv_uv",
    "Luv_to_LCHuv",
    "LCHuv_to_Luv",
]
__all__ += [
    "XYZ_to_UCS",
    "UCS_to_XYZ",
    "UCS_to_uv",
    "uv_to_UCS",
    "UCS_uv_to_xy",
    "xy_to_UCS_uv",
]
__all__ += [
    "XYZ_to_UVW",
    "UVW_to_XYZ",
]
__all__ += [
    "Lab_to_DIN99",
    "DIN99_to_Lab",
    "XYZ_to_DIN99",
    "DIN99_to_XYZ",
]
__all__ += [
    "HDR_CIELAB_METHODS",
    "XYZ_to_hdr_CIELab",
    "hdr_CIELab_to_XYZ",
]
__all__ += [
    "XYZ_to_K_ab_HunterLab1966",
    "XYZ_to_Hunter_Lab",
    "Hunter_Lab_to_XYZ",
    "XYZ_to_Hunter_Rdab",
]
__all__ += [
    "XYZ_to_Hunter_Rdab",
    "Hunter_Rdab_to_XYZ",
]
__all__ += [
    "XYZ_to_ICaCb",
    "ICaCb_to_XYZ",
]
__all__ += [
    "XYZ_to_IgPgTg",
    "IgPgTg_to_XYZ",
]
__all__ += [
    "XYZ_to_IPT",
    "IPT_to_XYZ",
    "IPT_hue_angle",
]
__all__ += [
    "IZAZBZ_METHODS",
    "XYZ_to_Izazbz",
    "XYZ_to_Jzazbz",
    "Izazbz_to_XYZ",
    "Jzazbz_to_XYZ",
]
__all__ += [
    "HDR_IPT_METHODS",
    "XYZ_to_hdr_IPT",
    "hdr_IPT_to_XYZ",
]
__all__ += [
    "XYZ_to_Oklab",
    "Oklab_to_XYZ",
]
__all__ += [
    "XYZ_to_OSA_UCS",
    "OSA_UCS_to_XYZ",
]
__all__ += [
    "XYZ_to_ProLab",
    "ProLab_to_XYZ",
]
__all__ += [
    "DATA_MACADAM_1942_ELLIPSES",
    "CCS_ILLUMINANT_POINTER_GAMUT",
    "DATA_POINTER_GAMUT_VOLUME",
    "CCS_POINTER_GAMUT_BOUNDARY",
]
__all__ += [
    "normalised_primary_matrix",
    "chromatically_adapted_primaries",
    "primaries_whitepoint",
    "RGB_luminance_equation",
    "RGB_luminance",
]
__all__ += ["RGB_Colourspace"]
__all__ += ["XYZ_to_RGB", "RGB_to_XYZ"]
__all__ += ["matrix_RGB_to_RGB", "RGB_to_RGB"]
__all__ += [
    "CV_range",
    "legal_to_full",
    "full_to_legal",
    "gamma_function",
    "log_encoding_ACESproxy",
    "log_decoding_ACESproxy",
    "log_encoding_ACEScc",
    "log_decoding_ACEScc",
    "log_encoding_ACEScct",
    "log_decoding_ACEScct",
    "oetf_ARIBSTDB67",
    "oetf_inverse_ARIBSTDB67",
    "log_encoding_ALEXALogC",
    "log_decoding_ALEXALogC",
    "oetf_BlackmagicFilmGeneration5",
    "oetf_inverse_BlackmagicFilmGeneration5",
    "log_encoding_CanonLog",
    "log_decoding_CanonLog",
    "log_encoding_CanonLog2",
    "log_decoding_CanonLog2",
    "log_encoding_CanonLog3",
    "log_decoding_CanonLog3",
    "log_encoding_Cineon",
    "log_decoding_Cineon",
    "oetf_DaVinciIntermediate",
    "oetf_inverse_DaVinciIntermediate",
    "eotf_inverse_DCDM",
    "eotf_DCDM",
    "eotf_inverse_DICOMGSDF",
    "eotf_DICOMGSDF",
    "log_encoding_DJIDLog",
    "log_decoding_DJIDLog",
    "exponent_function_basic",
    "exponent_function_monitor_curve",
    "log_encoding_FilmicPro6",
    "log_decoding_FilmicPro6",
    "log_encoding_FilmLightTLog",
    "log_decoding_FilmLightTLog",
    "log_encoding_Protune",
    "log_decoding_Protune",
    "oetf_BT601",
    "oetf_inverse_BT601",
    "oetf_BT709",
    "oetf_inverse_BT709",
    "eotf_inverse_BT1886",
    "eotf_BT1886",
    "eotf_inverse_BT2020",
    "eotf_BT2020",
    "eotf_inverse_ST2084",
    "eotf_ST2084",
    "oetf_PQ_BT2100",
    "oetf_inverse_PQ_BT2100",
    "eotf_PQ_BT2100",
    "eotf_inverse_PQ_BT2100",
    "ootf_PQ_BT2100",
    "ootf_inverse_PQ_BT2100",
    "oetf_HLG_BT2100",
    "oetf_inverse_HLG_BT2100",
    "BT2100_HLG_EOTF_METHODS",
    "eotf_HLG_BT2100",
    "BT2100_HLG_EOTF_INVERSE_METHODS",
    "eotf_inverse_HLG_BT2100",
    "BT2100_HLG_OOTF_METHODS",
    "ootf_HLG_BT2100",
    "BT2100_HLG_OOTF_INVERSE_METHODS",
    "ootf_inverse_HLG_BT2100",
    "linear_function",
    "logarithmic_function_basic",
    "logarithmic_function_quasilog",
    "logarithmic_function_camera",
    "log_encoding_Log2",
    "log_decoding_Log2",
    "log_encoding_Panalog",
    "log_decoding_Panalog",
    "log_encoding_VLog",
    "log_decoding_VLog",
    "log_encoding_FLog",
    "log_decoding_FLog",
    "log_encoding_NLog",
    "log_decoding_NLog",
    "log_encoding_PivotedLog",
    "log_decoding_PivotedLog",
    "log_encoding_REDLog",
    "log_decoding_REDLog",
    "log_encoding_REDLogFilm",
    "log_decoding_REDLogFilm",
    "LOG3G10_ENCODING_METHODS",
    "LOG3G10_DECODING_METHODS",
    "log_encoding_Log3G10",
    "log_decoding_Log3G10",
    "log_encoding_Log3G12",
    "log_decoding_Log3G12",
    "cctf_encoding_ROMMRGB",
    "cctf_decoding_ROMMRGB",
    "cctf_encoding_ProPhotoRGB",
    "cctf_decoding_ProPhotoRGB",
    "cctf_encoding_RIMMRGB",
    "cctf_decoding_RIMMRGB",
    "log_encoding_ERIMMRGB",
    "log_decoding_ERIMMRGB",
    "oetf_SMPTE240M",
    "eotf_SMPTE240M",
    "log_encoding_SLog",
    "log_decoding_SLog",
    "log_encoding_SLog2",
    "log_decoding_SLog2",
    "log_encoding_SLog3",
    "log_decoding_SLog3",
    "eotf_inverse_sRGB",
    "eotf_sRGB",
    "log_encoding_ViperLog",
    "log_decoding_ViperLog",
]
__all__ += [
    "LOG_ENCODINGS",
    "log_encoding",
    "LOG_DECODINGS",
    "log_decoding",
    "OETFS",
    "oetf",
    "OETF_INVERSES",
    "oetf_inverse",
    "EOTFS",
    "eotf",
    "EOTF_INVERSES",
    "eotf_inverse",
    "CCTF_ENCODINGS",
    "cctf_encoding",
    "CCTF_DECODINGS",
    "cctf_decoding",
    "OOTFS",
    "ootf",
    "OOTF_INVERSES",
    "ootf_inverse",
]
__all__ += [
    "RGB_COLOURSPACES",
    "RGB_COLOURSPACE_ACES2065_1",
    "RGB_COLOURSPACE_ACESCC",
    "RGB_COLOURSPACE_ACESCCT",
    "RGB_COLOURSPACE_ACESPROXY",
    "RGB_COLOURSPACE_ACESCG",
    "RGB_COLOURSPACE_ADOBE_RGB1998",
    "RGB_COLOURSPACE_ADOBE_WIDE_GAMUT_RGB",
    "RGB_COLOURSPACE_ALEXA_WIDE_GAMUT",
    "RGB_COLOURSPACE_APPLE_RGB",
    "RGB_COLOURSPACE_BEST_RGB",
    "RGB_COLOURSPACE_BETA_RGB",
    "RGB_COLOURSPACE_BLACKMAGIC_WIDE_GAMUT",
    "RGB_COLOURSPACE_BT470_525",
    "RGB_COLOURSPACE_BT470_625",
    "RGB_COLOURSPACE_BT709",
    "RGB_COLOURSPACE_BT2020",
    "RGB_COLOURSPACE_CIE_RGB",
    "RGB_COLOURSPACE_CINEMA_GAMUT",
    "RGB_COLOURSPACE_COLOR_MATCH_RGB",
    "RGB_COLOURSPACE_DAVINCI_WIDE_GAMUT",
    "RGB_COLOURSPACE_DCDM_XYZ",
    "RGB_COLOURSPACE_DCI_P3",
    "RGB_COLOURSPACE_DCI_P3_P",
    "RGB_COLOURSPACE_DISPLAY_P3",
    "RGB_COLOURSPACE_DJI_D_GAMUT",
    "RGB_COLOURSPACE_DON_RGB_4",
    "RGB_COLOURSPACE_ECI_RGB_V2",
    "RGB_COLOURSPACE_EKTA_SPACE_PS_5",
    "RGB_COLOURSPACE_FILMLIGHT_E_GAMUT",
    "RGB_COLOURSPACE_PROTUNE_NATIVE",
    "RGB_COLOURSPACE_MAX_RGB",
    "RGB_COLOURSPACE_N_GAMUT",
    "RGB_COLOURSPACE_P3_D65",
    "RGB_COLOURSPACE_PAL_SECAM",
    "RGB_COLOURSPACE_RED_COLOR",
    "RGB_COLOURSPACE_RED_COLOR_2",
    "RGB_COLOURSPACE_RED_COLOR_3",
    "RGB_COLOURSPACE_RED_COLOR_4",
    "RGB_COLOURSPACE_DRAGON_COLOR",
    "RGB_COLOURSPACE_DRAGON_COLOR_2",
    "RGB_COLOURSPACE_RED_WIDE_GAMUT_RGB",
    "RGB_COLOURSPACE_ROMM_RGB",
    "RGB_COLOURSPACE_RIMM_RGB",
    "RGB_COLOURSPACE_ERIMM_RGB",
    "RGB_COLOURSPACE_PROPHOTO_RGB",
    "RGB_COLOURSPACE_RUSSELL_RGB",
    "RGB_COLOURSPACE_SHARP_RGB",
    "RGB_COLOURSPACE_SMPTE_240M",
    "RGB_COLOURSPACE_SMPTE_C",
    "RGB_COLOURSPACE_NTSC1953",
    "RGB_COLOURSPACE_NTSC1987",
    "RGB_COLOURSPACE_S_GAMUT",
    "RGB_COLOURSPACE_S_GAMUT3",
    "RGB_COLOURSPACE_S_GAMUT3_CINE",
    "RGB_COLOURSPACE_VENICE_S_GAMUT3",
    "RGB_COLOURSPACE_VENICE_S_GAMUT3_CINE",
    "RGB_COLOURSPACE_sRGB",
    "RGB_COLOURSPACE_V_GAMUT",
    "RGB_COLOURSPACE_XTREME_RGB",
    "RGB_COLOURSPACE_F_GAMUT",
]

__all__ += ["XYZ_to_sRGB", "sRGB_to_XYZ"]
__all__ += [
    "RGB_to_HSV",
    "HSV_to_RGB",
    "RGB_to_HSL",
    "HSL_to_RGB",
    "RGB_to_HCL",
    "HCL_to_RGB",
]
__all__ += ["RGB_to_CMY", "CMY_to_RGB", "CMY_to_CMYK", "CMYK_to_CMY"]
__all__ += ["RGB_to_IHLS", "IHLS_to_RGB"]
__all__ += ["RGB_to_Prismatic", "Prismatic_to_RGB"]
__all__ += [
    "WEIGHTS_YCBCR",
    "matrix_YCbCr",
    "offset_YCbCr",
    "RGB_to_YCbCr",
    "YCbCr_to_RGB",
    "RGB_to_YcCbcCrc",
    "YcCbcCrc_to_RGB",
]
__all__ += ["RGB_to_YCoCg", "YCoCg_to_RGB"]
__all__ += ["RGB_to_ICtCp", "ICtCp_to_RGB", "XYZ_to_ICtCp", "ICtCp_to_XYZ"]


# ----------------------------------------------------------------------------#
# ---                API Changes and Deprecation Management                ---#
# ----------------------------------------------------------------------------#
class models(ModuleAPI):
    """Define a class acting like the *models* module."""

    def __getattr__(self, attribute) -> Any:
        """Return the value from the attribute with given name."""

        return super().__getattr__(attribute)


# v0.4.0
API_CHANGES = {
    "ObjectRenamed": [
        [
            "colour.models.RGB_to_ICTCP",
            "colour.models.RGB_to_ICtCp",
        ],
        [
            "colour.models.ICTCP_to_RGB",
            "colour.models.ICtCp_to_RGB",
        ],
        [
            "colour.models.RGB_to_IGPGTG",
            "colour.models.RGB_to_IgPgTg",
        ],
        [
            "colour.models.IGPGTG_to_RGB",
            "colour.models.IgPgTg_to_RGB",
        ],
        [
            "colour.models.XYZ_to_JzAzBz",
            "colour.models.XYZ_to_Jzazbz",
        ],
        [
            "colour.models.JzAzBz_to_XYZ",
            "colour.models.Jzazbz_to_XYZ",
        ],
    ]
}
"""Defines the *colour.models* sub-package API changes."""

if not is_documentation_building():
    sys.modules["colour.models"] = models(  # type: ignore[assignment]
        sys.modules["colour.models"], build_API_changes(API_CHANGES)
    )

    del ModuleAPI, is_documentation_building, build_API_changes, sys



"""Defines the unit tests for the :mod:`colour.plotting.models` module."""

import numpy as np
import unittest
from matplotlib.pyplot import Axes, Figure

from colour.plotting import (
    colourspace_model_axis_reorder,
    plot_pointer_gamut,
    plot_RGB_colourspaces_in_chromaticity_diagram_CIE1931,
    plot_RGB_colourspaces_in_chromaticity_diagram_CIE1960UCS,
    plot_RGB_colourspaces_in_chromaticity_diagram_CIE1976UCS,
    plot_RGB_chromaticities_in_chromaticity_diagram_CIE1931,
    plot_RGB_chromaticities_in_chromaticity_diagram_CIE1960UCS,
    plot_RGB_chromaticities_in_chromaticity_diagram_CIE1976UCS,
    plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1931,
    plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1960UCS,
    plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1976UCS,
    plot_single_cctf,
    plot_multi_cctfs,
    plot_constant_hue_loci,
)
from colour.plotting.models import (
    plot_RGB_colourspaces_in_chromaticity_diagram,
    plot_RGB_chromaticities_in_chromaticity_diagram,
    ellipses_MacAdam1942,
    plot_ellipses_MacAdam1942_in_chromaticity_diagram,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestCommonColourspaceModelAxisReorder",
    "TestPlotPointerGamut",
    "TestPlotRGBColourspacesInChromaticityDiagram",
    "TestPlotRGBColourspacesInChromaticityDiagramCIE1931",
    "TestPlotRGBColourspacesInChromaticityDiagramCIE1960UCS",
    "TestPlotRGBColourspacesInChromaticityDiagramCIE1976UCS",
    "TestPlotRGBChromaticitiesInChromaticityDiagram",
    "TestPlotRGBChromaticitiesInChromaticityDiagramCIE1931",
    "TestPlotRGBChromaticitiesInChromaticityDiagramCIE1960UCS",
    "TestPlotRGBChromaticitiesInChromaticityDiagramCIE1976UCS",
    "TestPlotEllipsesMacAdam1942InChromaticityDiagram",
    "TestPlotEllipsesMacAdam1942InChromaticityDiagramCIE1931",
    "TestPlotEllipsesMacAdam1942InChromaticityDiagramCIE1960UCS",
    "TestPlotEllipsesMacAdam1942InChromaticityDiagramCIE1976UCS",
    "TestPlotSingleCctf",
    "TestPlotMultiCctfs",
    "TestPlotConstantHueLoci",
]


class TestCommonColourspaceModelAxisReorder(unittest.TestCase):
    """
    Define :func:`colour.plotting.models.colourspace_model_axis_reorder`
    definition unit tests methods.
    """

    def test_colourspace_model_axis_reorder(self):
        """
        Test :func:`colour.plotting.models.colourspace_model_axis_reorder`
        definition.
        """

        a = np.array([0, 1, 2])

        np.testing.assert_almost_equal(
            colourspace_model_axis_reorder(a, "CIE Lab"),
            np.array([1, 2, 0]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            colourspace_model_axis_reorder(a, "IPT"),
            np.array([1, 2, 0]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            colourspace_model_axis_reorder(a, "OSA UCS"),
            np.array([1, 2, 0]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            colourspace_model_axis_reorder(
                colourspace_model_axis_reorder(a, "OSA UCS"),
                "OSA UCS",
                "Inverse",
            ),
            np.array([0, 1, 2]),
            decimal=7,
        )


class TestPlotPointerGamut(unittest.TestCase):
    """
    Define :func:`colour.plotting.models.plot_pointer_gamut` definition unit
    tests methods.
    """

    def test_plot_pointer_gamut(self):
        """Test :func:`colour.plotting.models.plot_pointer_gamut` definition."""

        figure, axes = plot_pointer_gamut()

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)

        figure, axes = plot_pointer_gamut(method="CIE 1960 UCS")

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)

        figure, axes = plot_pointer_gamut(method="CIE 1976 UCS")

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)

        self.assertRaises(
            ValueError, lambda: plot_pointer_gamut(method="Undefined")
        )


class TestPlotRGBColourspacesInChromaticityDiagram(unittest.TestCase):
    """
    Define :func:`colour.plotting.models.\
plot_RGB_colourspaces_in_chromaticity_diagram` definition unit tests methods.
    """

    def test_plot_RGB_colourspaces_in_chromaticity_diagram(self):
        """
        Test :func:`colour.plotting.models.\
plot_RGB_colourspaces_in_chromaticity_diagram` definition.
        """

        figure, axes = plot_RGB_colourspaces_in_chromaticity_diagram(
            ["ITU-R BT.709", "ACEScg", "S-Gamut"],
            show_pointer_gamut=True,
            chromatically_adapt=True,
            plot_kwargs={"linestyle": "dashed"},
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)

        figure, axes = plot_RGB_colourspaces_in_chromaticity_diagram(
            ["ITU-R BT.709", "ACEScg", "S-Gamut"],
            plot_kwargs=[{"linestyle": "dashed"}] * 3,
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)

        self.assertRaises(
            ValueError,
            lambda: plot_RGB_colourspaces_in_chromaticity_diagram(
                ["ITU-R BT.709", "ACEScg", "S-Gamut"],
                chromaticity_diagram_callable=lambda **x: x,
                method="Undefined",
            ),
        )


class TestPlotRGBColourspacesInChromaticityDiagramCIE1931(unittest.TestCase):
    """
    Define :func:`colour.plotting.models.\
plot_RGB_colourspaces_in_chromaticity_diagram_CIE1931` definition unit tests
    methods.
    """

    def test_plot_RGB_colourspaces_in_chromaticity_diagram_CIE1931(self):
        """
        Test :func:`colour.plotting.models.\
plot_RGB_colourspaces_in_chromaticity_diagram_CIE1931` definition.
        """

        figure, axes = plot_RGB_colourspaces_in_chromaticity_diagram_CIE1931(
            ["ITU-R BT.709", "ACEScg", "S-Gamut"]
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotRGBColourspacesInChromaticityDiagramCIE1960UCS(
    unittest.TestCase
):
    """
    Define :func:`colour.plotting.models.\
plot_RGB_colourspaces_in_chromaticity_diagram_CIE1960UCS` definition unit tests
    methods.
    """

    def test_plot_RGB_colourspaces_in_chromaticity_diagram_CIE1960UCS(self):
        """
        Test :func:`colour.plotting.models.\
plot_RGB_colourspaces_in_chromaticity_diagram_CIE1960UCS` definition.
        """

        (
            figure,
            axes,
        ) = plot_RGB_colourspaces_in_chromaticity_diagram_CIE1960UCS(
            ["ITU-R BT.709", "ACEScg", "S-Gamut"]
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotRGBColourspacesInChromaticityDiagramCIE1976UCS(
    unittest.TestCase
):
    """
    Define :func:`colour.plotting.models.\
plot_RGB_colourspaces_in_chromaticity_diagram_CIE1976UCS` definition unit tests
    methods.
    """

    def test_plot_RGB_colourspaces_in_chromaticity_diagram_CIE1976UCS(self):
        """
        Test :func:`colour.plotting.models.\
plot_RGB_colourspaces_in_chromaticity_diagram_CIE1976UCS` definition.
        """

        (
            figure,
            axes,
        ) = plot_RGB_colourspaces_in_chromaticity_diagram_CIE1976UCS(
            ["ITU-R BT.709", "ACEScg", "S-Gamut"]
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotRGBChromaticitiesInChromaticityDiagram(unittest.TestCase):
    """
    Define :func:`colour.plotting.models.\
plot_RGB_chromaticities_in_chromaticity_diagram` definition unit tests methods.
    """

    def test_plot_RGB_chromaticities_in_chromaticity_diagram(self):
        """
        Test :func:`colour.plotting.models.\
plot_RGB_chromaticities_in_chromaticity_diagram` definition.
        """

        figure, axes = plot_RGB_chromaticities_in_chromaticity_diagram(
            np.random.random((128, 128, 3)), scatter_kwargs={"marker": "v"}
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotRGBChromaticitiesInChromaticityDiagramCIE1931(unittest.TestCase):
    """
    Define :func:`colour.plotting.models.\
plot_RGB_chromaticities_in_chromaticity_diagram_CIE1931` definition unit tests
    methods.
    """

    def test_plot_RGB_chromaticities_in_chromaticity_diagram_CIE1931(self):
        """
        Test :func:`colour.plotting.models.\
plot_RGB_chromaticities_in_chromaticity_diagram_CIE1931` definition.
        """

        figure, axes = plot_RGB_chromaticities_in_chromaticity_diagram_CIE1931(
            np.random.random((128, 128, 3))
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotRGBChromaticitiesInChromaticityDiagramCIE1960UCS(
    unittest.TestCase
):
    """
    Define :func:`colour.plotting.models.\
plot_RGB_chromaticities_in_chromaticity_diagram_CIE1960UCS` definition unit
    tests methods.
    """

    def test_plot_RGB_chromaticities_in_chromaticity_diagram_CIE1960UCS(self):
        """
        Test :func:`colour.plotting.models.\
plot_RGB_chromaticities_in_chromaticity_diagram_CIE1960UCS` definition.
        """

        (
            figure,
            axes,
        ) = plot_RGB_chromaticities_in_chromaticity_diagram_CIE1960UCS(
            np.random.random((128, 128, 3))
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotRGBChromaticitiesInChromaticityDiagramCIE1976UCS(
    unittest.TestCase
):
    """
    Define :func:`colour.plotting.models.\
plot_RGB_chromaticities_in_chromaticity_diagram_CIE1976UCS` definition unit
    tests methods.
    """

    def test_plot_RGB_chromaticities_in_chromaticity_diagram_CIE1976UCS(self):
        """
        Test :func:`colour.plotting.models.\
plot_RGB_chromaticities_in_chromaticity_diagram_CIE1976UCS` definition.
        """

        (
            figure,
            axes,
        ) = plot_RGB_chromaticities_in_chromaticity_diagram_CIE1976UCS(
            np.random.random((128, 128, 3))
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestEllipsesMacAdam1942(unittest.TestCase):
    """
    Define :func:`colour.plotting.models.ellipses_MacAdam1942` definition unit
    tests methods.
    """

    def test_ellipses_MacAdam1942(self):
        """Test :func:`colour.plotting.models.ellipses_MacAdam1942` definition."""

        self.assertEqual(len(ellipses_MacAdam1942()), 25)

        self.assertRaises(
            ValueError, lambda: ellipses_MacAdam1942(method="Undefined")
        )


class TestPlotEllipsesMacAdam1942InChromaticityDiagram(unittest.TestCase):
    """
    Define :func:`colour.plotting.models.\
plot_ellipses_MacAdam1942_in_chromaticity_diagram` definition unit tests
    methods.
    """

    def test_plot_ellipses_MacAdam1942_in_chromaticity_diagram(self):
        """
        Test :func:`colour.plotting.models.\
plot_ellipses_MacAdam1942_in_chromaticity_diagram` definition.
        """

        figure, axes = plot_ellipses_MacAdam1942_in_chromaticity_diagram(
            chromaticity_diagram_clipping=True, ellipse_kwargs={"color": "k"}
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)

        figure, axes = plot_ellipses_MacAdam1942_in_chromaticity_diagram(
            chromaticity_diagram_clipping=True,
            ellipse_kwargs=[{"color": "k"}] * 25,
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotEllipsesMacAdam1942InChromaticityDiagramCIE1931(
    unittest.TestCase
):
    """
    Define :func:`colour.plotting.models.\
plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1931` definition unit
    tests methods.
    """

    def test_plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1931(self):
        """
        Test :func:`colour.plotting.models.\
plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1931` definition.
        """

        (
            figure,
            axes,
        ) = plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1931()

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotEllipsesMacAdam1942InChromaticityDiagramCIE1960UCS(
    unittest.TestCase
):
    """
    Define :func:`colour.plotting.models.\
plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1960UCS` definition unit
    tests methods.
    """

    def test_plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1960UCS(
        self,
    ):
        """
        Test :func:`colour.plotting.models.\
plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1960UCS` definition.
        """

        (
            figure,
            axes,
        ) = plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1960UCS()

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotEllipsesMacAdam1942InChromaticityDiagramCIE1976UCS(
    unittest.TestCase
):
    """
    Define :func:`colour.plotting.models.\
plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1976UCS` definition unit
    tests methods.
    """

    def test_plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1976UCS(
        self,
    ):
        """
        Test :func:`colour.plotting.models.\
plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1976UCS` definition.
        """

        (
            figure,
            axes,
        ) = plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1976UCS()

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotSingleCctf(unittest.TestCase):
    """
    Define :func:`colour.plotting.models.plot_single_cctf` definition unit
    tests methods.
    """

    def test_plot_single_cctf(self):
        """Test :func:`colour.plotting.models.plot_single_cctf` definition."""

        figure, axes = plot_single_cctf("ITU-R BT.709")

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotMultiCctfs(unittest.TestCase):
    """
    Define :func:`colour.plotting.models.plot_multi_cctfs` definition unit
    tests methods.
    """

    def test_plot_multi_cctfs(self):
        """Test :func:`colour.plotting.models.plot_multi_cctfs` definition."""

        figure, axes = plot_multi_cctfs(["ITU-R BT.709", "sRGB"])

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotConstantHueLoci(unittest.TestCase):
    """
    Define :func:`colour.plotting.models.plot_constant_hue_loci` definition
    unit tests methods.
    """

    def test_plot_constant_hue_loci(self):
        """Test :func:`colour.plotting.models.plot_constant_hue_loci` definition."""

        data = np.array(
            [
                [
                    None,
                    np.array([0.95010000, 1.00000000, 1.08810000]),
                    np.array([0.40920000, 0.28120000, 0.30600000]),
                    np.array(
                        [
                            [0.02495100, 0.01908600, 0.02032900],
                            [0.10944300, 0.06235900, 0.06788100],
                            [0.27186500, 0.18418700, 0.19565300],
                            [0.48898900, 0.40749400, 0.44854600],
                        ]
                    ),
                    None,
                ],
                [
                    None,
                    np.array([0.95010000, 1.00000000, 1.08810000]),
                    np.array([0.30760000, 0.48280000, 0.42770000]),
                    np.array(
                        [
                            [0.02108000, 0.02989100, 0.02790400],
                            [0.06194700, 0.11251000, 0.09334400],
                            [0.15255800, 0.28123300, 0.23234900],
                            [0.34157700, 0.56681300, 0.47035300],
                        ]
                    ),
                    None,
                ],
                [
                    None,
                    np.array([0.95010000, 1.00000000, 1.08810000]),
                    np.array([0.39530000, 0.28120000, 0.18450000]),
                    np.array(
                        [
                            [0.02436400, 0.01908600, 0.01468800],
                            [0.10331200, 0.06235900, 0.02854600],
                            [0.26311900, 0.18418700, 0.12109700],
                            [0.43158700, 0.40749400, 0.39008600],
                        ]
                    ),
                    None,
                ],
                [
                    None,
                    np.array([0.95010000, 1.00000000, 1.08810000]),
                    np.array([0.20510000, 0.18420000, 0.57130000]),
                    np.array(
                        [
                            [0.03039800, 0.02989100, 0.06123300],
                            [0.08870000, 0.08498400, 0.21843500],
                            [0.18405800, 0.18418700, 0.40111400],
                            [0.32550100, 0.34047200, 0.50296900],
                            [0.53826100, 0.56681300, 0.80010400],
                        ]
                    ),
                    None,
                ],
                [
                    None,
                    np.array([0.95010000, 1.00000000, 1.08810000]),
                    np.array([0.35770000, 0.28120000, 0.11250000]),
                    np.array(
                        [
                            [0.03678100, 0.02989100, 0.01481100],
                            [0.17127700, 0.11251000, 0.01229900],
                            [0.30080900, 0.28123300, 0.21229800],
                            [0.52976000, 0.40749400, 0.11720000],
                        ]
                    ),
                    None,
                ],
            ]
        )

        figure, axes = plot_constant_hue_loci(
            data, "IPT", scatter_kwargs={"marker": "v"}
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


if __name__ == "__main__":
    unittest.main()



"""
Rayleigh Optical Depth - Scattering in the Atmosphere
=====================================================

Implements *Rayleigh* scattering / optical depth in the atmosphere computation:

-   :func:`colour.scattering_cross_section`
-   :func:`colour.phenomena.rayleigh_optical_depth`
-   :func:`colour.rayleigh_scattering`

References
----------
-   :cite:`Bodhaine1999a` : Bodhaine, B. A., Wood, N. B., Dutton, E. G., &
    Slusser, J. R. (1999). On Rayleigh Optical Depth Calculations. Journal of
    Atmospheric and Oceanic Technology, 16(11), 1854-1861.
    doi:10.1175/1520-0426(1999)016<1854:ORODC>2.0.CO;2
-   :cite:`Wikipedia2001c` : Wikipedia. (2001). Rayleigh scattering. Retrieved
    September 23, 2014, from http://en.wikipedia.org/wiki/Rayleigh_scattering
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import (
    SPECTRAL_SHAPE_DEFAULT,
    SpectralDistribution,
    SpectralShape,
)
from colour.constants import CONSTANT_AVOGADRO
from colour.hints import Callable, FloatingOrArrayLike, FloatingOrNDArray
from colour.utilities import as_float, as_float_array, filter_kwargs

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "CONSTANT_STANDARD_AIR_TEMPERATURE",
    "CONSTANT_STANDARD_CO2_CONCENTRATION",
    "CONSTANT_AVERAGE_PRESSURE_MEAN_SEA_LEVEL",
    "CONSTANT_DEFAULT_LATITUDE",
    "CONSTANT_DEFAULT_ALTITUDE",
    "air_refraction_index_Penndorf1957",
    "air_refraction_index_Edlen1966",
    "air_refraction_index_Peck1972",
    "air_refraction_index_Bodhaine1999",
    "N2_depolarisation",
    "O2_depolarisation",
    "F_air_Penndorf1957",
    "F_air_Young1981",
    "F_air_Bates1984",
    "F_air_Bodhaine1999",
    "molecular_density",
    "mean_molecular_weights",
    "gravity_List1968",
    "scattering_cross_section",
    "rayleigh_optical_depth",
    "rayleigh_scattering",
]

CONSTANT_STANDARD_AIR_TEMPERATURE: float = 288.15
"""*Standard air* temperature :math:`T[K]` in kelvin degrees (:math:`15\\circ C`)."""

CONSTANT_STANDARD_CO2_CONCENTRATION: float = 300
"""*Standard air* :math:`CO_2` concentration in parts per million (ppm)."""

CONSTANT_AVERAGE_PRESSURE_MEAN_SEA_LEVEL: float = 101325
"""*Standard air* average pressure :math:`Hg` at mean sea-level in pascal (Pa)."""

CONSTANT_DEFAULT_LATITUDE: float = 0
"""Default latitude in degrees (equator)."""

CONSTANT_DEFAULT_ALTITUDE: float = 0
"""Default altitude in meters (sea level)."""


def air_refraction_index_Penndorf1957(
    wavelength: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the air refraction index :math:`n_s` from given wavelength
    :math:`\\lambda` in  micrometers (:math:`\\mu m`) using *Penndorf (1957)*
    method.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in micrometers (:math:`\\mu m`).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Air refraction index :math:`n_s`.

    Examples
    --------
    >>> air_refraction_index_Penndorf1957(0.555)  # doctest: +ELLIPSIS
    1.0002777...
    """

    wl = as_float_array(wavelength)

    n = 6432.8 + 2949810 / (146 - wl ** (-2)) + 25540 / (41 - wl ** (-2))
    n /= 1.0e8
    n += +1

    return n


def air_refraction_index_Edlen1966(
    wavelength: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the air refraction index :math:`n_s` from given wavelength
    :math:`\\lambda` in micrometers (:math:`\\mu m`) using *Edlen (1966)*
    method.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in micrometers (:math:`\\mu m`).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Air refraction index :math:`n_s`.

    Examples
    --------
    >>> air_refraction_index_Edlen1966(0.555)  # doctest: +ELLIPSIS
    1.0002777...
    """

    wl = as_float_array(wavelength)

    n = 8342.13 + 2406030 / (130 - wl ** (-2)) + 15997 / (38.9 - wl ** (-2))
    n /= 1.0e8
    n += +1

    return n


def air_refraction_index_Peck1972(
    wavelength: FloatingOrArrayLike,
) -> FloatingOrNDArray:
    """
    Return the air refraction index :math:`n_s` from given wavelength
    :math:`\\lambda` in micrometers (:math:`\\mu m`) using
    *Peck and Reeder (1972)* method.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in micrometers (:math:`\\mu m`).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Air refraction index :math:`n_s`.

    Examples
    --------
    >>> air_refraction_index_Peck1972(0.555)  # doctest: +ELLIPSIS
    1.0002777...
    """

    wl = as_float_array(wavelength)

    n = (
        8060.51
        + 2480990 / (132.274 - wl ** (-2))
        + 17455.7 / (39.32957 - wl ** (-2))
    )
    n /= 1.0e8
    n += +1

    return n


def air_refraction_index_Bodhaine1999(
    wavelength: FloatingOrArrayLike,
    CO2_concentration: FloatingOrArrayLike = CONSTANT_STANDARD_CO2_CONCENTRATION,
) -> FloatingOrNDArray:
    """
    Return the air refraction index :math:`n_s` from given wavelength
    :math:`\\lambda` in micrometers (:math:`\\mu m`) using
    *Bodhaine, Wood, Dutton and Slusser (1999)* method.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in micrometers (:math:`\\mu m`).
    CO2_concentration
        :math:`CO_2` concentration in parts per million (ppm).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Air refraction index :math:`n_s`.

    Examples
    --------
    >>> air_refraction_index_Bodhaine1999(0.555)  # doctest: +ELLIPSIS
    1.0002777...
    """

    wl = as_float_array(wavelength)
    CO2_c = as_float_array(CO2_concentration)

    n = (1 + 0.54 * ((CO2_c * 1e-6) - 300e-6)) * (
        air_refraction_index_Peck1972(wl) - 1
    ) + 1

    return as_float(n)


def N2_depolarisation(wavelength: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the depolarisation of nitrogen :math:`N_2` as function of
    wavelength :math:`\\lambda` in micrometers (:math:`\\mu m`).

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in micrometers (:math:`\\mu m`).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Nitrogen :math:`N_2` depolarisation.

    Examples
    --------
    >>> N2_depolarisation(0.555)  # doctest: +ELLIPSIS
    1.0350291...
    """

    wl = as_float_array(wavelength)

    N2 = 1.034 + 3.17 * 1.0e-4 * (1 / wl**2)

    return N2


def O2_depolarisation(wavelength: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return the depolarisation of oxygen :math:`O_2` as function of
    wavelength :math:`\\lambda` in micrometers (:math:`\\mu m`).

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in micrometers (:math:`\\mu m`).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Oxygen :math:`O_2` depolarisation.

    Examples
    --------
    >>> O2_depolarisation(0.555)  # doctest: +ELLIPSIS
    1.1020225...
    """

    wl = as_float_array(wavelength)

    O2 = (
        1.096 + 1.385 * 1.0e-3 * (1 / wl**2) + 1.448 * 1.0e-4 * (1 / wl**4)
    )

    return O2


def F_air_Penndorf1957(wavelength: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return :math:`(6+3_p)/(6-7_p)`, the depolarisation term :math:`F(air)` or
    *King Factor* using *Penndorf (1957)* method.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in micrometers (:math:`\\mu m`).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Air depolarisation.

    Notes
    -----
    -   The argument *wavelength* is only provided for consistency with the
        other air depolarisation methods but is actually not used as this
        definition is essentially a constant in its current implementation.

    Examples
    --------
    >>> F_air_Penndorf1957(0.555)
    1.0608
    """

    wl = as_float_array(wavelength)

    return as_float(np.resize(np.array([1.0608]), wl.shape))


def F_air_Young1981(wavelength: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return :math:`(6+3_p)/(6-7_p)`, the depolarisation term :math:`F(air)` or
    *King Factor* using *Young (1981)* method.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in micrometers (:math:`\\mu m`).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Air depolarisation.

    Notes
    -----
    -   The argument *wavelength* is only provided for consistency with the
        other air depolarisation methods but is actually not used as this
        definition is essentially a constant in its current implementation.

    Examples
    --------
    >>> F_air_Young1981(0.555)
    1.048
    """

    wl = as_float_array(wavelength)

    return as_float(np.resize(np.array([1.0480]), wl.shape))


def F_air_Bates1984(wavelength: FloatingOrArrayLike) -> FloatingOrNDArray:
    """
    Return :math:`(6+3_p)/(6-7_p)`, the depolarisation term :math:`F(air)` or
    *King Factor* as function of wavelength :math:`\\lambda` in micrometers
    (:math:`\\mu m`) using *Bates (1984)* method.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in micrometers (:math:`\\mu m`).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Air depolarisation.

    Examples
    --------
    >>> F_air_Bates1984(0.555)  # doctest: +ELLIPSIS
    1.0481535...
    """

    O2 = O2_depolarisation(wavelength)
    N2 = N2_depolarisation(wavelength)
    Ar = 1.00
    CO2 = 1.15

    F_air = (78.084 * N2 + 20.946 * O2 + CO2 + Ar) / (
        78.084 + 20.946 + Ar + CO2
    )

    return F_air


def F_air_Bodhaine1999(
    wavelength: FloatingOrArrayLike,
    CO2_concentration: FloatingOrArrayLike = CONSTANT_STANDARD_CO2_CONCENTRATION,
) -> FloatingOrNDArray:
    """
    Return :math:`(6+3_p)/(6-7_p)`, the depolarisation term :math:`F(air)` or
    *King Factor* as function of wavelength :math:`\\lambda` in micrometers
    (:math:`\\mu m`) and :math:`CO_2` concentration in parts per million (ppm)
    using *Bodhaine, Wood, Dutton and Slusser (1999)* method.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in micrometers (:math:`\\mu m`).
    CO2_concentration
        :math:`CO_2` concentration in parts per million (ppm).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Air depolarisation.

    Examples
    --------
    >>> F_air_Bodhaine1999(0.555)  # doctest: +ELLIPSIS
    1.1246916...
    """

    O2 = O2_depolarisation(wavelength)
    N2 = N2_depolarisation(wavelength)
    CO2_c = as_float_array(CO2_concentration)

    F_air = (78.084 * N2 + 20.946 * O2 + 0.934 * 1 + CO2_c * 1.15) / (
        78.084 + 20.946 + 0.934 + CO2_c
    )

    return F_air


def molecular_density(
    temperature: FloatingOrArrayLike = CONSTANT_STANDARD_AIR_TEMPERATURE,
    avogadro_constant: FloatingOrArrayLike = CONSTANT_AVOGADRO,
) -> FloatingOrNDArray:
    """
    Return the molecular density :math:`N_s` (molecules :math:`cm^{-3}`)
    as function of air temperature :math:`T[K]` in kelvin degrees.

    Parameters
    ----------
    temperature
        Air temperature :math:`T[K]` in kelvin degrees.
    avogadro_constant
        *Avogadro*'s number (molecules :math:`mol^{-1}`).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Molecular density :math:`N_s` (molecules :math:`cm^{-3}`).

    Notes
    -----
    -   The *Avogadro*'s number used in this implementation is the one given by
        by the Committee on Data for Science and Technology (CODATA):
        :math:`6.02214179x10^{23}`, which is different from the reference
        :cite:`Bodhaine1999a` value :math:`6.0221367x10^{23}`.

    Examples
    --------
    >>> molecular_density(288.15)  # doctest: +ELLIPSIS
    2.5469021...e+19
    >>> molecular_density(288.15, 6.0221367e23)  # doctest: +ELLIPSIS
    2.5468999...e+19
    """

    T = as_float_array(temperature)
    avogadro_constant = as_float_array(avogadro_constant)

    N_s = (avogadro_constant / 22.4141) * (273.15 / T) * (1 / 1000)

    return N_s


def mean_molecular_weights(
    CO2_concentration: FloatingOrArrayLike = CONSTANT_STANDARD_CO2_CONCENTRATION,
) -> FloatingOrNDArray:
    """
    Return the mean molecular weights :math:`m_a` for dry air as function of
    :math:`CO_2` concentration in parts per million (ppm).

    Parameters
    ----------
    CO2_concentration
        :math:`CO_2` concentration in parts per million (ppm).

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Mean molecular weights :math:`m_a` for dry air.

    Examples
    --------
    >>> mean_molecular_weights()  # doctest: +ELLIPSIS
    28.9640166...
    """

    CO2_concentration = as_float_array(CO2_concentration)

    CO2_c = CO2_concentration * 1.0e-6

    m_a = 15.0556 * CO2_c + 28.9595
    return m_a


def gravity_List1968(
    latitude: FloatingOrArrayLike = CONSTANT_DEFAULT_LATITUDE,
    altitude: FloatingOrArrayLike = CONSTANT_DEFAULT_ALTITUDE,
) -> FloatingOrNDArray:
    """
    Return the gravity :math:`g` in :math:`cm/s_2` (gal) representative of the
    mass-weighted column of air molecules above the site of given latitude and
    altitude using *List (1968)* method.

    Parameters
    ----------
    latitude
        Latitude of the site in degrees.
    altitude
        Altitude of the site in meters.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Gravity :math:`g` in :math:`cm/s_2` (gal).

    Examples
    --------
    >>> gravity_List1968()  # doctest: +ELLIPSIS
    978.0356070...
    >>> gravity_List1968(0.0, 1500.0)  # doctest: +ELLIPSIS
    977.5726106...

    Gravity :math:`g` for Paris:

    >>> gravity_List1968(48.8567, 35.0)  # doctest: +ELLIPSIS
    980.9524178...
    """

    latitude = as_float_array(latitude)
    altitude = as_float_array(altitude)

    cos2phi = np.cos(2 * np.radians(latitude))

    # Sea level acceleration of gravity.
    g0 = 980.6160 * (1 - 0.0026373 * cos2phi + 0.0000059 * cos2phi**2)

    g = (
        g0
        - (3.085462e-4 + 2.27e-7 * cos2phi) * altitude
        + (7.254e-11 + 1.0e-13 * cos2phi) * altitude**2
        - (1.517e-17 + 6e-20 * cos2phi) * altitude**3
    )

    return g


def scattering_cross_section(
    wavelength: FloatingOrArrayLike,
    CO2_concentration: FloatingOrArrayLike = CONSTANT_STANDARD_CO2_CONCENTRATION,
    temperature: FloatingOrArrayLike = CONSTANT_STANDARD_AIR_TEMPERATURE,
    avogadro_constant: FloatingOrArrayLike = CONSTANT_AVOGADRO,
    n_s_function: Callable = air_refraction_index_Bodhaine1999,
    F_air_function: Callable = F_air_Bodhaine1999,
) -> FloatingOrNDArray:
    """
    Return the scattering cross section per molecule :math:`\\sigma` of dry
    air as function of wavelength :math:`\\lambda` in centimeters (cm) using
    given :math:`CO_2` concentration in parts per million (ppm) and temperature
    :math:`T[K]` in kelvin degrees following *Van de Hulst (1957)* method.

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in centimeters (cm).
    CO2_concentration
        :math:`CO_2` concentration in parts per million (ppm).
    temperature
        Air temperature :math:`T[K]` in kelvin degrees.
    avogadro_constant
        *Avogadro*'s number (molecules :math:`mol^{-1}`).
    n_s_function
        Air refraction index :math:`n_s` computation method.
    F_air_function
        :math:`(6+3_p)/(6-7_p)`, the depolarisation term :math:`F(air)` or
        *King Factor* computation method.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        Scattering cross section per molecule :math:`\\sigma` of dry air.

    Warnings
    --------
    Unlike most objects of :mod:`colour.phenomena.rayleigh` module,
    :func:`colour.scattering_cross_section` expects wavelength :math:`\\lambda`
    to be expressed in centimeters (cm).

    References
    ----------
    :cite:`Bodhaine1999a`, :cite:`Wikipedia2001c`

    Examples
    --------
    >>> scattering_cross_section(555 * 10e-8)  # doctest: +ELLIPSIS
    4.6613309...e-27
    """

    wl = as_float_array(wavelength)
    CO2_c = as_float_array(CO2_concentration)
    temperature = as_float_array(temperature)

    wl_micrometers = wl * 10e3

    N_s = molecular_density(temperature, avogadro_constant)
    n_s = n_s_function(wl_micrometers)
    # n_s = n_s_function(**filter_kwargs(
    #     n_s_function, wavelength=wl_micrometers, CO2_concentration=CO2_c))
    F_air = F_air_function(
        **filter_kwargs(
            F_air_function, wavelength=wl_micrometers, CO2_concentration=CO2_c
        )
    )

    sigma = (
        24
        * np.pi**3
        * (n_s**2 - 1) ** 2
        / (wl**4 * N_s**2 * (n_s**2 + 2) ** 2)
    )
    sigma *= F_air

    return sigma


def rayleigh_optical_depth(
    wavelength: FloatingOrArrayLike,
    CO2_concentration: FloatingOrArrayLike = CONSTANT_STANDARD_CO2_CONCENTRATION,
    temperature: FloatingOrArrayLike = CONSTANT_STANDARD_AIR_TEMPERATURE,
    pressure: FloatingOrArrayLike = CONSTANT_AVERAGE_PRESSURE_MEAN_SEA_LEVEL,
    latitude: FloatingOrArrayLike = CONSTANT_DEFAULT_LATITUDE,
    altitude: FloatingOrArrayLike = CONSTANT_DEFAULT_ALTITUDE,
    avogadro_constant: FloatingOrArrayLike = CONSTANT_AVOGADRO,
    n_s_function: Callable = air_refraction_index_Bodhaine1999,
    F_air_function: Callable = F_air_Bodhaine1999,
) -> FloatingOrNDArray:
    """
    Return the *Rayleigh* optical depth :math:`T_r(\\lambda)` as function of
    wavelength :math:`\\lambda` in centimeters (cm).

    Parameters
    ----------
    wavelength
        Wavelength :math:`\\lambda` in centimeters (cm).
    CO2_concentration
        :math:`CO_2` concentration in parts per million (ppm).
    temperature
        Air temperature :math:`T[K]` in kelvin degrees.
    pressure
        Surface pressure :math:`P` of the measurement site.
    latitude
        Latitude of the site in degrees.
    altitude
        Altitude of the site in meters.
    avogadro_constant
        *Avogadro*'s number (molecules :math:`mol^{-1}`).
    n_s_function
        Air refraction index :math:`n_s` computation method.
    F_air_function
        :math:`(6+3_p)/(6-7_p)`, the depolarisation term :math:`F(air)` or
        *King Factor* computation method.

    Returns
    -------
    :class:`numpy.floating` or :class:`numpy.ndarray`
        *Rayleigh* optical depth :math:`T_r(\\lambda)`.

    Warnings
    --------
    Unlike most objects of :mod:`colour.phenomena.rayleigh` module,
    :func:`colour.phenomena.rayleigh_optical_depth` expects wavelength
    :math:`\\lambda` to be expressed in centimeters (cm).

    References
    ----------
    :cite:`Bodhaine1999a`, :cite:`Wikipedia2001c`

    Examples
    --------
    >>> rayleigh_optical_depth(555 * 10e-8)  # doctest: +ELLIPSIS
    0.1004070...
    """

    wavelength = as_float_array(wavelength)
    CO2_c = as_float_array(CO2_concentration)
    pressure = as_float_array(pressure)
    latitude = as_float_array(latitude)
    altitude = as_float_array(altitude)
    # Conversion from pascal to dyne/cm2.
    P = as_float_array(pressure * 10)

    sigma = scattering_cross_section(
        wavelength,
        CO2_c,
        temperature,
        avogadro_constant,
        n_s_function,
        F_air_function,
    )

    m_a = mean_molecular_weights(CO2_c)
    g = gravity_List1968(latitude, altitude)

    T_R = sigma * (P * avogadro_constant) / (m_a * g)

    return as_float(T_R)


rayleigh_scattering = rayleigh_optical_depth


def sd_rayleigh_scattering(
    shape: SpectralShape = SPECTRAL_SHAPE_DEFAULT,
    CO2_concentration: FloatingOrArrayLike = CONSTANT_STANDARD_CO2_CONCENTRATION,
    temperature: FloatingOrArrayLike = CONSTANT_STANDARD_AIR_TEMPERATURE,
    pressure: FloatingOrArrayLike = CONSTANT_AVERAGE_PRESSURE_MEAN_SEA_LEVEL,
    latitude: FloatingOrArrayLike = CONSTANT_DEFAULT_LATITUDE,
    altitude: FloatingOrArrayLike = CONSTANT_DEFAULT_ALTITUDE,
    avogadro_constant: FloatingOrArrayLike = CONSTANT_AVOGADRO,
    n_s_function: Callable = air_refraction_index_Bodhaine1999,
    F_air_function: Callable = F_air_Bodhaine1999,
) -> SpectralDistribution:
    """
    Return the *Rayleigh* spectral distribution for given spectral shape.

    Parameters
    ----------
    shape
        Spectral shape used to create the *Rayleigh* scattering spectral
        distribution.
    CO2_concentration
        :math:`CO_2` concentration in parts per million (ppm).
    temperature
        Air temperature :math:`T[K]` in kelvin degrees.
    pressure
        Surface pressure :math:`P` of the measurement site.
    latitude
        Latitude of the site in degrees.
    altitude
        Altitude of the site in meters.
    avogadro_constant
        *Avogadro*'s number (molecules :math:`mol^{-1}`).
    n_s_function
        Air refraction index :math:`n_s` computation method.
    F_air_function
        :math:`(6+3_p)/(6-7_p)`, the depolarisation term :math:`F(air)` or
        *King Factor* computation method.

    Returns
    -------
    :class:`colour.SpectralDistribution`
        *Rayleigh* optical depth spectral distribution.

    References
    ----------
    :cite:`Bodhaine1999a`, :cite:`Wikipedia2001c`

    Examples
    --------
    >>> from colour.utilities import numpy_print_options
    >>> with numpy_print_options(suppress=True):
    ...     sd_rayleigh_scattering()  # doctest: +ELLIPSIS
    SpectralDistribution([[ 360.        ,    0.5991013...],
                          [ 361.        ,    0.5921706...],
                          [ 362.        ,    0.5853410...],
                          [ 363.        ,    0.5786105...],
                          [ 364.        ,    0.5719774...],
                          [ 365.        ,    0.5654401...],
                          [ 366.        ,    0.5589968...],
                          [ 367.        ,    0.5526460...],
                          [ 368.        ,    0.5463860...],
                          [ 369.        ,    0.5402153...],
                          [ 370.        ,    0.5341322...],
                          [ 371.        ,    0.5281354...],
                          [ 372.        ,    0.5222234...],
                          [ 373.        ,    0.5163946...],
                          [ 374.        ,    0.5106476...],
                          [ 375.        ,    0.5049812...],
                          [ 376.        ,    0.4993939...],
                          [ 377.        ,    0.4938844...],
                          [ 378.        ,    0.4884513...],
                          [ 379.        ,    0.4830934...],
                          [ 380.        ,    0.4778095...],
                          [ 381.        ,    0.4725983...],
                          [ 382.        ,    0.4674585...],
                          [ 383.        ,    0.4623891...],
                          [ 384.        ,    0.4573889...],
                          [ 385.        ,    0.4524566...],
                          [ 386.        ,    0.4475912...],
                          [ 387.        ,    0.4427917...],
                          [ 388.        ,    0.4380568...],
                          [ 389.        ,    0.4333856...],
                          [ 390.        ,    0.4287771...],
                          [ 391.        ,    0.4242302...],
                          [ 392.        ,    0.4197439...],
                          [ 393.        ,    0.4153172...],
                          [ 394.        ,    0.4109493...],
                          [ 395.        ,    0.4066391...],
                          [ 396.        ,    0.4023857...],
                          [ 397.        ,    0.3981882...],
                          [ 398.        ,    0.3940458...],
                          [ 399.        ,    0.3899576...],
                          [ 400.        ,    0.3859227...],
                          [ 401.        ,    0.3819402...],
                          [ 402.        ,    0.3780094...],
                          [ 403.        ,    0.3741295...],
                          [ 404.        ,    0.3702996...],
                          [ 405.        ,    0.366519 ...],
                          [ 406.        ,    0.3627868...],
                          [ 407.        ,    0.3591025...],
                          [ 408.        ,    0.3554651...],
                          [ 409.        ,    0.3518740...],
                          [ 410.        ,    0.3483286...],
                          [ 411.        ,    0.344828 ...],
                          [ 412.        ,    0.3413716...],
                          [ 413.        ,    0.3379587...],
                          [ 414.        ,    0.3345887...],
                          [ 415.        ,    0.3312609...],
                          [ 416.        ,    0.3279747...],
                          [ 417.        ,    0.3247294...],
                          [ 418.        ,    0.3215245...],
                          [ 419.        ,    0.3183593...],
                          [ 420.        ,    0.3152332...],
                          [ 421.        ,    0.3121457...],
                          [ 422.        ,    0.3090962...],
                          [ 423.        ,    0.3060841...],
                          [ 424.        ,    0.3031088...],
                          [ 425.        ,    0.3001699...],
                          [ 426.        ,    0.2972668...],
                          [ 427.        ,    0.2943989...],
                          [ 428.        ,    0.2915657...],
                          [ 429.        ,    0.2887668...],
                          [ 430.        ,    0.2860017...],
                          [ 431.        ,    0.2832697...],
                          [ 432.        ,    0.2805706...],
                          [ 433.        ,    0.2779037...],
                          [ 434.        ,    0.2752687...],
                          [ 435.        ,    0.2726650...],
                          [ 436.        ,    0.2700922...],
                          [ 437.        ,    0.2675500...],
                          [ 438.        ,    0.2650377...],
                          [ 439.        ,    0.2625551...],
                          [ 440.        ,    0.2601016...],
                          [ 441.        ,    0.2576770...],
                          [ 442.        ,    0.2552807...],
                          [ 443.        ,    0.2529124...],
                          [ 444.        ,    0.2505716...],
                          [ 445.        ,    0.2482581...],
                          [ 446.        ,    0.2459713...],
                          [ 447.        ,    0.2437110...],
                          [ 448.        ,    0.2414768...],
                          [ 449.        ,    0.2392683...],
                          [ 450.        ,    0.2370851...],
                          [ 451.        ,    0.2349269...],
                          [ 452.        ,    0.2327933...],
                          [ 453.        ,    0.2306841...],
                          [ 454.        ,    0.2285989...],
                          [ 455.        ,    0.2265373...],
                          [ 456.        ,    0.2244990...],
                          [ 457.        ,    0.2224838...],
                          [ 458.        ,    0.2204912...],
                          [ 459.        ,    0.2185211...],
                          [ 460.        ,    0.2165730...],
                          [ 461.        ,    0.2146467...],
                          [ 462.        ,    0.2127419...],
                          [ 463.        ,    0.2108583...],
                          [ 464.        ,    0.2089957...],
                          [ 465.        ,    0.2071536...],
                          [ 466.        ,    0.2053320...],
                          [ 467.        ,    0.2035304...],
                          [ 468.        ,    0.2017487...],
                          [ 469.        ,    0.1999865...],
                          [ 470.        ,    0.1982436...],
                          [ 471.        ,    0.1965198...],
                          [ 472.        ,    0.1948148...],
                          [ 473.        ,    0.1931284...],
                          [ 474.        ,    0.1914602...],
                          [ 475.        ,    0.1898101...],
                          [ 476.        ,    0.1881779...],
                          [ 477.        ,    0.1865633...],
                          [ 478.        ,    0.1849660...],
                          [ 479.        ,    0.1833859...],
                          [ 480.        ,    0.1818227...],
                          [ 481.        ,    0.1802762...],
                          [ 482.        ,    0.1787463...],
                          [ 483.        ,    0.1772326...],
                          [ 484.        ,    0.1757349...],
                          [ 485.        ,    0.1742532...],
                          [ 486.        ,    0.1727871...],
                          [ 487.        ,    0.1713365...],
                          [ 488.        ,    0.1699011...],
                          [ 489.        ,    0.1684809...],
                          [ 490.        ,    0.1670755...],
                          [ 491.        ,    0.1656848...],
                          [ 492.        ,    0.1643086...],
                          [ 493.        ,    0.1629468...],
                          [ 494.        ,    0.1615991...],
                          [ 495.        ,    0.1602654...],
                          [ 496.        ,    0.1589455...],
                          [ 497.        ,    0.1576392...],
                          [ 498.        ,    0.1563464...],
                          [ 499.        ,    0.1550668...],
                          [ 500.        ,    0.1538004...],
                          [ 501.        ,    0.1525470...],
                          [ 502.        ,    0.1513063...],
                          [ 503.        ,    0.1500783...],
                          [ 504.        ,    0.1488628...],
                          [ 505.        ,    0.1476597...],
                          [ 506.        ,    0.1464687...],
                          [ 507.        ,    0.1452898...],
                          [ 508.        ,    0.1441228...],
                          [ 509.        ,    0.1429675...],
                          [ 510.        ,    0.1418238...],
                          [ 511.        ,    0.1406916...],
                          [ 512.        ,    0.1395707...],
                          [ 513.        ,    0.1384610...],
                          [ 514.        ,    0.1373624...],
                          [ 515.        ,    0.1362747...],
                          [ 516.        ,    0.1351978...],
                          [ 517.        ,    0.1341316...],
                          [ 518.        ,    0.1330759...],
                          [ 519.        ,    0.1320306...],
                          [ 520.        ,    0.1309956...],
                          [ 521.        ,    0.1299707...],
                          [ 522.        ,    0.1289559...],
                          [ 523.        ,    0.1279511...],
                          [ 524.        ,    0.1269560...],
                          [ 525.        ,    0.1259707...],
                          [ 526.        ,    0.1249949...],
                          [ 527.        ,    0.1240286...],
                          [ 528.        ,    0.1230717...],
                          [ 529.        ,    0.1221240...],
                          [ 530.        ,    0.1211855...],
                          [ 531.        ,    0.1202560...],
                          [ 532.        ,    0.1193354...],
                          [ 533.        ,    0.1184237...],
                          [ 534.        ,    0.1175207...],
                          [ 535.        ,    0.1166263...],
                          [ 536.        ,    0.1157404...],
                          [ 537.        ,    0.1148630...],
                          [ 538.        ,    0.1139939...],
                          [ 539.        ,    0.1131331...],
                          [ 540.        ,    0.1122804...],
                          [ 541.        ,    0.1114357...],
                          [ 542.        ,    0.1105990...],
                          [ 543.        ,    0.1097702...],
                          [ 544.        ,    0.1089492...],
                          [ 545.        ,    0.1081358...],
                          [ 546.        ,    0.1073301...],
                          [ 547.        ,    0.1065319...],
                          [ 548.        ,    0.1057411...],
                          [ 549.        ,    0.1049577...],
                          [ 550.        ,    0.1041815...],
                          [ 551.        ,    0.1034125...],
                          [ 552.        ,    0.1026507...],
                          [ 553.        ,    0.1018958...],
                          [ 554.        ,    0.1011480...],
                          [ 555.        ,    0.1004070...],
                          [ 556.        ,    0.0996728...],
                          [ 557.        ,    0.0989453...],
                          [ 558.        ,    0.0982245...],
                          [ 559.        ,    0.0975102...],
                          [ 560.        ,    0.0968025...],
                          [ 561.        ,    0.0961012...],
                          [ 562.        ,    0.0954062...],
                          [ 563.        ,    0.0947176...],
                          [ 564.        ,    0.0940352...],
                          [ 565.        ,    0.0933589...],
                          [ 566.        ,    0.0926887...],
                          [ 567.        ,    0.0920246...],
                          [ 568.        ,    0.0913664...],
                          [ 569.        ,    0.0907141...],
                          [ 570.        ,    0.0900677...],
                          [ 571.        ,    0.0894270...],
                          [ 572.        ,    0.0887920...],
                          [ 573.        ,    0.0881627...],
                          [ 574.        ,    0.0875389...],
                          [ 575.        ,    0.0869207...],
                          [ 576.        ,    0.0863079...],
                          [ 577.        ,    0.0857006...],
                          [ 578.        ,    0.0850986...],
                          [ 579.        ,    0.0845019...],
                          [ 580.        ,    0.0839104...],
                          [ 581.        ,    0.0833242...],
                          [ 582.        ,    0.0827430...],
                          [ 583.        ,    0.082167 ...],
                          [ 584.        ,    0.0815959...],
                          [ 585.        ,    0.0810298...],
                          [ 586.        ,    0.0804687...],
                          [ 587.        ,    0.0799124...],
                          [ 588.        ,    0.0793609...],
                          [ 589.        ,    0.0788142...],
                          [ 590.        ,    0.0782722...],
                          [ 591.        ,    0.0777349...],
                          [ 592.        ,    0.0772022...],
                          [ 593.        ,    0.0766740...],
                          [ 594.        ,    0.0761504...],
                          [ 595.        ,    0.0756313...],
                          [ 596.        ,    0.0751166...],
                          [ 597.        ,    0.0746063...],
                          [ 598.        ,    0.0741003...],
                          [ 599.        ,    0.0735986...],
                          [ 600.        ,    0.0731012...],
                          [ 601.        ,    0.072608 ...],
                          [ 602.        ,    0.0721189...],
                          [ 603.        ,    0.0716340...],
                          [ 604.        ,    0.0711531...],
                          [ 605.        ,    0.0706763...],
                          [ 606.        ,    0.0702035...],
                          [ 607.        ,    0.0697347...],
                          [ 608.        ,    0.0692697...],
                          [ 609.        ,    0.0688087...],
                          [ 610.        ,    0.0683515...],
                          [ 611.        ,    0.0678981...],
                          [ 612.        ,    0.0674485...],
                          [ 613.        ,    0.0670026...],
                          [ 614.        ,    0.0665603...],
                          [ 615.        ,    0.0661218...],
                          [ 616.        ,    0.0656868...],
                          [ 617.        ,    0.0652555...],
                          [ 618.        ,    0.0648277...],
                          [ 619.        ,    0.0644033...],
                          [ 620.        ,    0.0639825...],
                          [ 621.        ,    0.0635651...],
                          [ 622.        ,    0.0631512...],
                          [ 623.        ,    0.0627406...],
                          [ 624.        ,    0.0623333...],
                          [ 625.        ,    0.0619293...],
                          [ 626.        ,    0.0615287...],
                          [ 627.        ,    0.0611312...],
                          [ 628.        ,    0.0607370...],
                          [ 629.        ,    0.0603460...],
                          [ 630.        ,    0.0599581...],
                          [ 631.        ,    0.0595733...],
                          [ 632.        ,    0.0591917...],
                          [ 633.        ,    0.0588131...],
                          [ 634.        ,    0.0584375...],
                          [ 635.        ,    0.0580649...],
                          [ 636.        ,    0.0576953...],
                          [ 637.        ,    0.0573286...],
                          [ 638.        ,    0.0569649...],
                          [ 639.        ,    0.0566040...],
                          [ 640.        ,    0.0562460...],
                          [ 641.        ,    0.0558909...],
                          [ 642.        ,    0.0555385...],
                          [ 643.        ,    0.0551890...],
                          [ 644.        ,    0.0548421...],
                          [ 645.        ,    0.0544981...],
                          [ 646.        ,    0.0541567...],
                          [ 647.        ,    0.053818 ...],
                          [ 648.        ,    0.0534819...],
                          [ 649.        ,    0.0531485...],
                          [ 650.        ,    0.0528176...],
                          [ 651.        ,    0.0524894...],
                          [ 652.        ,    0.0521637...],
                          [ 653.        ,    0.0518405...],
                          [ 654.        ,    0.0515198...],
                          [ 655.        ,    0.0512017...],
                          [ 656.        ,    0.0508859...],
                          [ 657.        ,    0.0505726...],
                          [ 658.        ,    0.0502618...],
                          [ 659.        ,    0.0499533...],
                          [ 660.        ,    0.0496472...],
                          [ 661.        ,    0.0493434...],
                          [ 662.        ,    0.0490420...],
                          [ 663.        ,    0.0487428...],
                          [ 664.        ,    0.0484460...],
                          [ 665.        ,    0.0481514...],
                          [ 666.        ,    0.0478591...],
                          [ 667.        ,    0.0475689...],
                          [ 668.        ,    0.0472810...],
                          [ 669.        ,    0.0469953...],
                          [ 670.        ,    0.0467117...],
                          [ 671.        ,    0.0464302...],
                          [ 672.        ,    0.0461509...],
                          [ 673.        ,    0.0458737...],
                          [ 674.        ,    0.0455986...],
                          [ 675.        ,    0.0453255...],
                          [ 676.        ,    0.0450545...],
                          [ 677.        ,    0.0447855...],
                          [ 678.        ,    0.0445185...],
                          [ 679.        ,    0.0442535...],
                          [ 680.        ,    0.0439905...],
                          [ 681.        ,    0.0437294...],
                          [ 682.        ,    0.0434703...],
                          [ 683.        ,    0.0432131...],
                          [ 684.        ,    0.0429578...],
                          [ 685.        ,    0.0427044...],
                          [ 686.        ,    0.0424529...],
                          [ 687.        ,    0.0422032...],
                          [ 688.        ,    0.0419553...],
                          [ 689.        ,    0.0417093...],
                          [ 690.        ,    0.0414651...],
                          [ 691.        ,    0.0412226...],
                          [ 692.        ,    0.0409820...],
                          [ 693.        ,    0.0407431...],
                          [ 694.        ,    0.0405059...],
                          [ 695.        ,    0.0402705...],
                          [ 696.        ,    0.0400368...],
                          [ 697.        ,    0.0398047...],
                          [ 698.        ,    0.0395744...],
                          [ 699.        ,    0.0393457...],
                          [ 700.        ,    0.0391187...],
                          [ 701.        ,    0.0388933...],
                          [ 702.        ,    0.0386696...],
                          [ 703.        ,    0.0384474...],
                          [ 704.        ,    0.0382269...],
                          [ 705.        ,    0.0380079...],
                          [ 706.        ,    0.0377905...],
                          [ 707.        ,    0.0375747...],
                          [ 708.        ,    0.0373604...],
                          [ 709.        ,    0.0371476...],
                          [ 710.        ,    0.0369364...],
                          [ 711.        ,    0.0367266...],
                          [ 712.        ,    0.0365184...],
                          [ 713.        ,    0.0363116...],
                          [ 714.        ,    0.0361063...],
                          [ 715.        ,    0.0359024...],
                          [ 716.        ,    0.0357000...],
                          [ 717.        ,    0.0354990...],
                          [ 718.        ,    0.0352994...],
                          [ 719.        ,    0.0351012...],
                          [ 720.        ,    0.0349044...],
                          [ 721.        ,    0.0347090...],
                          [ 722.        ,    0.0345150...],
                          [ 723.        ,    0.0343223...],
                          [ 724.        ,    0.0341310...],
                          [ 725.        ,    0.0339410...],
                          [ 726.        ,    0.0337523...],
                          [ 727.        ,    0.033565 ...],
                          [ 728.        ,    0.0333789...],
                          [ 729.        ,    0.0331941...],
                          [ 730.        ,    0.0330106...],
                          [ 731.        ,    0.0328284...],
                          [ 732.        ,    0.0326474...],
                          [ 733.        ,    0.0324677...],
                          [ 734.        ,    0.0322893...],
                          [ 735.        ,    0.0321120...],
                          [ 736.        ,    0.0319360...],
                          [ 737.        ,    0.0317611...],
                          [ 738.        ,    0.0315875...],
                          [ 739.        ,    0.0314151...],
                          [ 740.        ,    0.0312438...],
                          [ 741.        ,    0.0310737...],
                          [ 742.        ,    0.0309048...],
                          [ 743.        ,    0.0307370...],
                          [ 744.        ,    0.0305703...],
                          [ 745.        ,    0.0304048...],
                          [ 746.        ,    0.0302404...],
                          [ 747.        ,    0.0300771...],
                          [ 748.        ,    0.0299149...],
                          [ 749.        ,    0.0297538...],
                          [ 750.        ,    0.0295938...],
                          [ 751.        ,    0.0294349...],
                          [ 752.        ,    0.0292771...],
                          [ 753.        ,    0.0291203...],
                          [ 754.        ,    0.0289645...],
                          [ 755.        ,    0.0288098...],
                          [ 756.        ,    0.0286561...],
                          [ 757.        ,    0.0285035...],
                          [ 758.        ,    0.0283518...],
                          [ 759.        ,    0.0282012...],
                          [ 760.        ,    0.0280516...],
                          [ 761.        ,    0.0279030...],
                          [ 762.        ,    0.0277553...],
                          [ 763.        ,    0.0276086...],
                          [ 764.        ,    0.027463 ...],
                          [ 765.        ,    0.0273182...],
                          [ 766.        ,    0.0271744...],
                          [ 767.        ,    0.0270316...],
                          [ 768.        ,    0.0268897...],
                          [ 769.        ,    0.0267487...],
                          [ 770.        ,    0.0266087...],
                          [ 771.        ,    0.0264696...],
                          [ 772.        ,    0.0263314...],
                          [ 773.        ,    0.0261941...],
                          [ 774.        ,    0.0260576...],
                          [ 775.        ,    0.0259221...],
                          [ 776.        ,    0.0257875...],
                          [ 777.        ,    0.0256537...],
                          [ 778.        ,    0.0255208...],
                          [ 779.        ,    0.0253888...],
                          [ 780.        ,    0.0252576...]],
                         interpolator=SpragueInterpolator,
                         interpolator_kwargs={},
                         extrapolator=Extrapolator,
                         extrapolator_kwargs={...})
    """

    wavelengths = shape.range()
    return SpectralDistribution(
        rayleigh_optical_depth(
            wavelengths * 10e-8,
            CO2_concentration,
            temperature,
            pressure,
            latitude,
            altitude,
            avogadro_constant,
            n_s_function,
            F_air_function,
        ),
        wavelengths,
        name=(
            "Rayleigh Scattering - "
            f"{CO2_concentration!r} ppm, "
            f"{temperature!r} K, "
            f"{pressure!r} Pa, "
            f"{latitude!r} Degrees, "
            f"{altitude!r} m"
        ),
    )



"""Defines the unit tests for the :mod:`colour.recovery.mallett2019` module."""

import unittest
import numpy as np

from colour.characterisation import SDS_COLOURCHECKERS
from colour.colorimetry import (
    MSDS_CMFS,
    CCS_ILLUMINANTS,
    SDS_ILLUMINANTS,
    SpectralShape,
    reshape_msds,
    reshape_sd,
    sd_to_XYZ,
)
from colour.difference import JND_CIE1976, delta_E_CIE1976
from colour.models import (
    RGB_COLOURSPACE_PAL_SECAM,
    RGB_COLOURSPACE_sRGB,
    XYZ_to_RGB,
    XYZ_to_Lab,
)
from colour.recovery import (
    MSDS_BASIS_FUNCTIONS_sRGB_MALLETT2019,
    spectral_primary_decomposition_Mallett2019,
    RGB_to_sd_Mallett2019,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestMixinMallett2019",
    "TestSpectralPrimaryDecompositionMallett2019",
    "TestRGB_to_sd_Mallett2019",
]


class TestMixinMallett2019:
    """A mixin for testing the :mod:`colour.recovery.mallett2019` module."""

    def __init__(self):
        """Initialise common tests attributes for the mixin."""

        # pylint: disable=E1102
        self._cmfs = reshape_msds(
            MSDS_CMFS["CIE 1931 2 Degree Standard Observer"],
            SpectralShape(360, 780, 10),
        )
        self._sd_D65 = reshape_sd(SDS_ILLUMINANTS["D65"], self._cmfs.shape)
        self._xy_D65 = CCS_ILLUMINANTS["CIE 1931 2 Degree Standard Observer"][
            "D65"
        ]

    def check_basis_functions(self):
        """
        Test :func:`colour.recovery.RGB_to_sd_Mallett2019` definition or the
        more specialised :func:`colour.recovery.RGB_to_sd_Mallett2019`
        definition.
        """

        # Make sure the white point is reconstructed as a perfectly flat
        # spectrum.
        RGB = np.full(3, 1.0)
        sd = RGB_to_sd_Mallett2019(RGB, self._basis)
        self.assertLess(np.var(sd.values), 1e-5)

        # Check if the primaries or their combination exceeds the [0, 1] range.
        lower = np.zeros_like(sd.values) - 1e-12
        upper = np.ones_like(sd.values) + 1e12
        for RGB in [[1, 1, 1], [1, 0, 0], [0, 1, 0], [0, 0, 1]]:
            sd = RGB_to_sd_Mallett2019(RGB, self._basis)
            np.testing.assert_array_less(sd.values, upper)
            np.testing.assert_array_less(lower, sd.values)

        # Check Delta E's using a colour checker.
        for name, sd in SDS_COLOURCHECKERS["ColorChecker N Ohta"].items():
            XYZ = sd_to_XYZ(sd, self._cmfs, self._sd_D65) / 100
            Lab = XYZ_to_Lab(XYZ, self._xy_D65)
            RGB = XYZ_to_RGB(
                XYZ,
                self._RGB_colourspace.whitepoint,
                self._xy_D65,
                self._RGB_colourspace.matrix_XYZ_to_RGB,
            )

            recovered_sd = RGB_to_sd_Mallett2019(RGB, self._basis)
            recovered_XYZ = (
                sd_to_XYZ(recovered_sd, self._cmfs, self._sd_D65) / 100
            )
            recovered_Lab = XYZ_to_Lab(recovered_XYZ, self._xy_D65)

            error = delta_E_CIE1976(Lab, recovered_Lab)

            if error > 4 * JND_CIE1976 / 100:  # pragma: no cover
                self.fail(f'Delta E for "{name}" is {error}!')


class TestSpectralPrimaryDecompositionMallett2019(
    unittest.TestCase, TestMixinMallett2019
):
    """
    Define :func:`colour.recovery.spectral_primary_decomposition_Mallett2019`
    definition unit tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        TestMixinMallett2019.__init__(self)

        self._RGB_colourspace = RGB_COLOURSPACE_PAL_SECAM

    def test_spectral_primary_decomposition_Mallett2019(self):
        """
        Test :func:`colour.recovery.\
test_spectral_primary_decomposition_Mallett2019` definition.
        """

        self._basis = spectral_primary_decomposition_Mallett2019(
            self._RGB_colourspace, self._cmfs, self._sd_D65
        )

        self.check_basis_functions()


class TestRGB_to_sd_Mallett2019(unittest.TestCase, TestMixinMallett2019):
    """
    Define :func:`colour.recovery.RGB_to_sd_Mallett2019` definition unit
    tests methods.
    """

    def setUp(self):
        """Initialise the common tests attributes."""

        TestMixinMallett2019.__init__(self)

        self._RGB_colourspace = RGB_COLOURSPACE_sRGB
        self._basis = MSDS_BASIS_FUNCTIONS_sRGB_MALLETT2019

    def test_RGB_to_sd_Mallett2019(self):
        """Test :func:`colour.recovery.RGB_to_sd_Mallett2019` definition."""

        self.check_basis_functions()


if __name__ == "__main__":
    unittest.main()



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour.appearance.cam16` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.appearance import (
    VIEWING_CONDITIONS_CAM16,
    InductionFactors_CAM16,
    CAM_Specification_CAM16,
    XYZ_to_CAM16,
    CAM16_to_XYZ,
)
from colour.utilities import (
    as_float_array,
    domain_range_scale,
    ignore_numpy_errors,
    tsplit,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestXYZ_to_CAM16",
    "TestCAM16_to_XYZ",
]


class TestXYZ_to_CAM16(unittest.TestCase):
    """
    Define :func:`colour.appearance.cam16.XYZ_to_CAM16` definition unit
    tests methods.
    """

    def test_XYZ_to_CAM16(self):
        """Test :func:`colour.appearance.cam16.XYZ_to_CAM16` definition."""

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_w = np.array([95.05, 100.00, 108.88])
        L_A = 318.31
        Y_b = 20
        surround = VIEWING_CONDITIONS_CAM16["Average"]
        np.testing.assert_almost_equal(
            XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround),
            np.array(
                [
                    41.73120791,
                    0.10335574,
                    217.06795977,
                    2.34501507,
                    195.37170899,
                    0.10743677,
                    275.59498615,
                    np.nan,
                ]
            ),
            decimal=7,
        )

        XYZ = np.array([57.06, 43.06, 31.96])
        L_A = 31.83
        np.testing.assert_almost_equal(
            XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround),
            np.array(
                [
                    65.42828069,
                    49.67956420,
                    17.48659243,
                    52.94308868,
                    152.06985268,
                    42.62473321,
                    398.03047943,
                    np.nan,
                ]
            ),
            decimal=7,
        )

        XYZ = np.array([3.53, 6.56, 2.14])
        XYZ_w = np.array([109.85, 100, 35.58])
        L_A = 318.31
        np.testing.assert_almost_equal(
            XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround),
            np.array(
                [
                    21.36052893,
                    50.99381895,
                    178.86724266,
                    61.57953092,
                    139.78582768,
                    53.00732582,
                    223.01823806,
                    np.nan,
                ]
            ),
            decimal=7,
        )

        XYZ = np.array([19.01, 20.00, 21.78])
        L_A = 318.31
        np.testing.assert_almost_equal(
            XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround),
            np.array(
                [
                    41.36326063,
                    52.81154022,
                    258.88676291,
                    53.12406914,
                    194.52011798,
                    54.89682038,
                    311.24768647,
                    np.nan,
                ]
            ),
            decimal=7,
        )

        XYZ = np.array([61.45276998, 7.00421901, 82.2406738])
        XYZ_w = np.array([95.05, 100.00, 108.88])
        L_A = 4.074366543152521
        np.testing.assert_almost_equal(
            XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround),
            np.array(
                [
                    21.03801957,
                    457.78881613,
                    350.06445098,
                    241.50642846,
                    56.74143988,
                    330.94646237,
                    376.43915877,
                    np.nan,
                ]
            ),
            decimal=7,
        )

    def test_n_dimensional_XYZ_to_CAM16(self):
        """
        Test :func:`colour.appearance.cam16.XYZ_to_CAM16` definition
        n-dimensional support.
        """

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_w = np.array([95.05, 100.00, 108.88])
        L_A = 318.31
        Y_b = 20
        surround = VIEWING_CONDITIONS_CAM16["Average"]
        specification = XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround)

        XYZ = np.tile(XYZ, (6, 1))
        specification = np.tile(specification, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround),
            specification,
            decimal=7,
        )

        XYZ_w = np.tile(XYZ_w, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround),
            specification,
            decimal=7,
        )

        XYZ = np.reshape(XYZ, (2, 3, 3))
        XYZ_w = np.reshape(XYZ_w, (2, 3, 3))
        specification = np.reshape(specification, (2, 3, 8))
        np.testing.assert_almost_equal(
            XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround),
            specification,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_domain_range_scale_XYZ_to_CAM16(self):
        """
        Test :func:`colour.appearance.cam16.XYZ_to_CAM16` definition domain
        and range scale support.
        """

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_w = np.array([95.05, 100.00, 108.88])
        L_A = 318.31
        Y_b = 20
        surround = VIEWING_CONDITIONS_CAM16["Average"]
        specification = XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround)

        d_r = (
            ("reference", 1, 1),
            (
                "1",
                0.01,
                np.array(
                    [
                        1 / 100,
                        1 / 100,
                        1 / 360,
                        1 / 100,
                        1 / 100,
                        1 / 100,
                        1 / 400,
                        np.nan,
                    ]
                ),
            ),
            (
                "100",
                1,
                np.array([1, 1, 100 / 360, 1, 1, 1, 100 / 400, np.nan]),
            ),
        )
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    XYZ_to_CAM16(
                        XYZ * factor_a, XYZ_w * factor_a, L_A, Y_b, surround
                    ),
                    as_float_array(specification) * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_XYZ_to_CAM16(self):
        """
        Test :func:`colour.appearance.cam16.XYZ_to_CAM16` definition
        nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ = np.array(case)
            XYZ_w = np.array(case)
            L_A = case[0]
            Y_b = case[0]
            surround = InductionFactors_CAM16(case[0], case[0], case[0])
            XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround)


class TestCAM16_to_XYZ(unittest.TestCase):
    """
    Define :func:`colour.appearance.cam16.CAM16_to_XYZ` definition unit tests
    methods.
    """

    def test_CAM16_to_XYZ(self):
        """Test :func:`colour.appearance.cam16.CAM16_to_XYZ` definition."""

        specification = CAM_Specification_CAM16(
            41.73120791, 0.10335574, 217.06795977
        )
        XYZ_w = np.array([95.05, 100.00, 108.88])
        L_A = 318.31
        Y_b = 20
        surround = VIEWING_CONDITIONS_CAM16["Average"]
        np.testing.assert_almost_equal(
            CAM16_to_XYZ(specification, XYZ_w, L_A, Y_b, surround),
            np.array([19.01, 20.00, 21.78]),
            decimal=7,
        )

        specification = CAM_Specification_CAM16(
            65.42828069, 49.67956420, 17.48659243
        )
        L_A = 31.83
        np.testing.assert_almost_equal(
            CAM16_to_XYZ(specification, XYZ_w, L_A, Y_b, surround),
            np.array([57.06, 43.06, 31.96]),
            decimal=7,
        )

        specification = CAM_Specification_CAM16(
            21.36052893, 50.99381895, 178.86724266
        )
        XYZ_w = np.array([109.85, 100, 35.58])
        L_A = 318.31
        np.testing.assert_almost_equal(
            CAM16_to_XYZ(specification, XYZ_w, L_A, Y_b, surround),
            np.array([3.53, 6.56, 2.14]),
            decimal=7,
        )

        specification = CAM_Specification_CAM16(
            41.36326063, 52.81154022, 258.88676291
        )
        L_A = 318.31
        np.testing.assert_almost_equal(
            CAM16_to_XYZ(specification, XYZ_w, L_A, Y_b, surround),
            np.array([19.01, 20.00, 21.78]),
            decimal=7,
        )

        specification = CAM_Specification_CAM16(
            21.03801957, 457.78881613, 350.06445098
        )
        XYZ_w = np.array([95.05, 100.00, 108.88])
        L_A = 4.074366543152521
        np.testing.assert_almost_equal(
            CAM16_to_XYZ(specification, XYZ_w, L_A, Y_b, surround),
            np.array([61.45276998, 7.00421901, 82.2406738]),
            decimal=7,
        )

    def test_n_dimensional_CAM16_to_XYZ(self):
        """
        Test :func:`colour.appearance.cam16.CAM16_to_XYZ` definition
        n-dimensional support.
        """

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_w = np.array([95.05, 100.00, 108.88])
        L_A = 318.31
        Y_b = 20
        surround = VIEWING_CONDITIONS_CAM16["Average"]
        specification = XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround)
        XYZ = CAM16_to_XYZ(specification, XYZ_w, L_A, Y_b, surround)

        specification = CAM_Specification_CAM16(
            *np.transpose(np.tile(tsplit(specification), (6, 1))).tolist()
        )
        XYZ = np.tile(XYZ, (6, 1))
        np.testing.assert_almost_equal(
            CAM16_to_XYZ(specification, XYZ_w, L_A, Y_b, surround),
            XYZ,
            decimal=7,
        )

        XYZ_w = np.tile(XYZ_w, (6, 1))
        np.testing.assert_almost_equal(
            CAM16_to_XYZ(specification, XYZ_w, L_A, Y_b, surround),
            XYZ,
            decimal=7,
        )

        specification = CAM_Specification_CAM16(
            *tsplit(np.reshape(specification, (2, 3, 8))).tolist()
        )
        XYZ_w = np.reshape(XYZ_w, (2, 3, 3))
        XYZ = np.reshape(XYZ, (2, 3, 3))
        np.testing.assert_almost_equal(
            CAM16_to_XYZ(specification, XYZ_w, L_A, Y_b, surround),
            XYZ,
            decimal=7,
        )

    @ignore_numpy_errors
    def test_domain_range_scale_CAM16_to_XYZ(self):
        """
        Test :func:`colour.appearance.cam16.CAM16_to_XYZ` definition domain
        and range scale support.
        """

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_w = np.array([95.05, 100.00, 108.88])
        L_A = 318.31
        Y_b = 20
        surround = VIEWING_CONDITIONS_CAM16["Average"]
        specification = XYZ_to_CAM16(XYZ, XYZ_w, L_A, Y_b, surround)
        XYZ = CAM16_to_XYZ(specification, XYZ_w, L_A, Y_b, surround)

        d_r = (
            ("reference", 1, 1),
            (
                "1",
                np.array(
                    [
                        1 / 100,
                        1 / 100,
                        1 / 360,
                        1 / 100,
                        1 / 100,
                        1 / 100,
                        1 / 400,
                        np.nan,
                    ]
                ),
                0.01,
            ),
            (
                "100",
                np.array([1, 1, 100 / 360, 1, 1, 1, 100 / 400, np.nan]),
                1,
            ),
        )
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    CAM16_to_XYZ(
                        specification * factor_a,
                        XYZ_w * factor_b,
                        L_A,
                        Y_b,
                        surround,
                    ),
                    XYZ * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_raise_exception_CAM16_to_XYZ(self):
        """
        Test :func:`colour.appearance.cam16.CAM16_to_XYZ` definition raised
        exception.
        """

        self.assertRaises(
            ValueError,
            CAM16_to_XYZ,
            CAM_Specification_CAM16(
                41.731207905126638, None, 217.06795976739301
            ),
            np.array([95.05, 100.00, 108.88]),
            318.31,
            20.0,
            VIEWING_CONDITIONS_CAM16["Average"],
        )

    @ignore_numpy_errors
    def test_nan_CAM16_to_XYZ(self):
        """
        Test :func:`colour.appearance.cam16.CAM16_to_XYZ` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            J = case[0]
            C = case[0]
            h = case[0]
            XYZ_w = np.array(case)
            L_A = case[0]
            Y_b = case[0]
            surround = InductionFactors_CAM16(case[0], case[0], case[0])
            CAM16_to_XYZ(
                CAM_Specification_CAM16(J, C, h, M=50),
                XYZ_w,
                L_A,
                Y_b,
                surround,
            )



from .chromaticity_coordinates import CCS_COLOURCHECKERS, ColourChecker
from .sds import SDS_COLOURCHECKERS

__all__ = [
    "CCS_COLOURCHECKERS",
    "ColourChecker",
]
__all__ += [
    "SDS_COLOURCHECKERS",
]



"""
SMPTE 240M Colourspace
======================

Defines the *SMPTE 240M* colourspace:

-   :attr:`colour.models.RGB_COLOURSPACE_SMPTE_240M`.

References
----------
-   :cite:`SocietyofMotionPictureandTelevisionEngineers1999b` : Society of
    Motion Picture and Television Engineers. (1999). ANSI/SMPTE 240M-1995 -
    Signal Parameters - 1125-Line High-Definition Production Systems (pp. 1-7).
    http://car.france3.mars.free.fr/HD/INA-%2026%20jan%2006/\
SMPTE%20normes%20et%20confs/s240m.pdf
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import (
    RGB_Colourspace,
    normalised_primary_matrix,
    oetf_SMPTE240M,
    eotf_SMPTE240M,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_SMPTE_240M",
    "WHITEPOINT_NAME_SMPTE_240M",
    "CCS_WHITEPOINT_SMPTE_240M",
    "MATRIX_SMPTE_240M_TO_XYZ",
    "MATRIX_XYZ_TO_SMPTE_240M",
    "RGB_COLOURSPACE_SMPTE_240M",
]

PRIMARIES_SMPTE_240M: NDArray = np.array(
    [
        [0.6300, 0.3400],
        [0.3100, 0.5950],
        [0.1550, 0.0700],
    ]
)
"""*SMPTE 240M* colourspace primaries."""

WHITEPOINT_NAME_SMPTE_240M: str = "D65"
"""*SMPTE 240M* colourspace whitepoint name."""

CCS_WHITEPOINT_SMPTE_240M: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_SMPTE_240M]
"""*SMPTE 240M* colourspace whitepoint chromaticity coordinates."""

MATRIX_SMPTE_240M_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_SMPTE_240M, CCS_WHITEPOINT_SMPTE_240M
)
"""*SMPTE 240M* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_SMPTE_240M: NDArray = np.linalg.inv(MATRIX_SMPTE_240M_TO_XYZ)
"""*CIE XYZ* tristimulus values to *SMPTE 240M* colourspace matrix."""

RGB_COLOURSPACE_SMPTE_240M: RGB_Colourspace = RGB_Colourspace(
    "SMPTE 240M",
    PRIMARIES_SMPTE_240M,
    CCS_WHITEPOINT_SMPTE_240M,
    WHITEPOINT_NAME_SMPTE_240M,
    MATRIX_SMPTE_240M_TO_XYZ,
    MATRIX_XYZ_TO_SMPTE_240M,
    oetf_SMPTE240M,
    eotf_SMPTE240M,
)
RGB_COLOURSPACE_SMPTE_240M.__doc__ = """
*SMPTE 240M* colourspace.

References
----------
:cite:`SocietyofMotionPictureandTelevisionEngineers1999b`,
"""



"""Defines the unit tests for the :mod:`colour.plotting.notation` module."""

import unittest
from matplotlib.pyplot import Axes, Figure

from colour.plotting import (
    plot_single_munsell_value_function,
    plot_multi_munsell_value_functions,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestPlotSingleMunsellValueFunction",
    "TestPlotMultiMunsellValueFunctions",
]


class TestPlotSingleMunsellValueFunction(unittest.TestCase):
    """
    Define :func:`colour.plotting.notation.plot_single_munsell_value_function`
    definition unit tests methods.
    """

    def test_plot_single_munsell_value_function(self):
        """
        Test :func:`colour.plotting.notation.\
plot_single_munsell_value_function` definition.
        """

        figure, axes = plot_single_munsell_value_function("ASTM D1535")

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


class TestPlotMultiMunsellValueFunctions(unittest.TestCase):
    """
    Define :func:`colour.plotting.notation.plot_multi_munsell_value_functions`
    definition unit tests methods.
    """

    def test_plot_multi_munsell_value_functions(self):
        """
        Test :func:`colour.plotting.notation.\
plot_multi_munsell_value_functions` definition.
        """

        figure, axes = plot_multi_munsell_value_functions(
            ["ASTM D1535", "McCamy 1987"]
        )

        self.assertIsInstance(figure, Figure)
        self.assertIsInstance(axes, Axes)


if __name__ == "__main__":
    unittest.main()



"""
Cinespace .csp LUT Format Input / Output Utilities
==================================================

Defines the *Cinespace* *.csp* *LUT* format related input / output utilities
objects:

-   :func:`colour.io.read_LUT_Cinespace`
-   :func:`colour.io.write_LUT_Cinespace`

References
----------
-   :cite:`RisingSunResearch` : Rising Sun Research. (n.d.). cineSpace LUT
    Library. Retrieved November 30, 2018, from
    https://sourceforge.net/projects/cinespacelutlib/
"""

from __future__ import annotations

import numpy as np

from colour.hints import (
    ArrayLike,
    Boolean,
    Integer,
    List,
    NDArray,
    Tuple,
    Union,
)
from colour.io.luts import LUT1D, LUT3x1D, LUT3D, LUTSequence
from colour.utilities import (
    attest,
    tsplit,
    tstack,
    as_float_array,
    as_int_array,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "read_LUT_Cinespace",
    "write_LUT_Cinespace",
]


def read_LUT_Cinespace(path: str) -> Union[LUT3x1D, LUT3D, LUTSequence]:
    """
    Read given *Cinespace* *.csp* *LUT* file.

    Parameters
    ----------
    path
        *LUT* path.

    Returns
    -------
    :class:`colour.LUT3x1D` or :class:`colour.LUT3D` or \
:class:`colour.LUTSequence`
        :class:`LUT3x1D` or :class:`LUT3D` or :class:`LUTSequence` class
        instance.

    References
    ----------
    :cite:`RisingSunResearch`

    Examples
    --------
    Reading a 3x1D *Cinespace* *.csp* *LUT*:

    >>> import os
    >>> path = os.path.join(
    ...     os.path.dirname(__file__), 'tests', 'resources', 'cinespace',
    ...     'ACES_Proxy_10_to_ACES.csp')
    >>> print(read_LUT_Cinespace(path))
    LUT3x1D - ACES Proxy 10 to ACES
    -------------------------------
    <BLANKLINE>
    Dimensions : 2
    Domain     : [[ 0.  0.  0.]
                  [ 1.  1.  1.]]
    Size       : (32, 3)

    Reading a 3D *Cinespace* *.csp* *LUT*:

    >>> path = os.path.join(
    ...     os.path.dirname(__file__), 'tests', 'resources', 'cinespace',
    ...     'Colour_Correct.csp')
    >>> print(read_LUT_Cinespace(path))
    LUT3D - Generated by Foundry::LUT
    ---------------------------------
    <BLANKLINE>
    Dimensions : 3
    Domain     : [[ 0.  0.  0.]
                  [ 1.  1.  1.]]
    Size       : (4, 4, 4, 3)
    """

    unity_range = np.array([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])

    def _parse_metadata_section(metadata: List) -> Tuple:
        """Parse the metadata at given lines."""

        return (metadata[0], metadata[1:]) if len(metadata) > 0 else ("", [])

    def _parse_domain_section(lines: List[str]) -> NDArray:
        """Parse the domain at given lines."""

        pre_LUT_size = max(int(lines[i]) for i in [0, 3, 6])
        pre_LUT = [
            as_float_array(lines[i].split()) for i in [1, 2, 4, 5, 7, 8]
        ]

        pre_LUT_padded = []
        for row in pre_LUT:
            if len(row) != pre_LUT_size:
                pre_LUT_padded.append(
                    np.pad(
                        row,
                        (0, pre_LUT_size - row.shape[0]),
                        mode="constant",
                        constant_values=np.nan,
                    )
                )
            else:
                pre_LUT_padded.append(row)

        return np.asarray(pre_LUT_padded)

    def _parse_table_section(lines):
        """Parse the table at given lines."""

        size = as_int_array(lines[0].split())
        table = as_float_array([line.split() for line in lines[1:]])

        return size, table

    with open(path) as csp_file:
        lines = csp_file.readlines()
        attest(len(lines) > 0, '"LUT" is empty!')
        lines = [line.strip() for line in lines if line.strip()]

        header = lines[0]
        attest(header == "CSPLUTV100", '"LUT" header is invalid!')

        kind = lines[1]
        attest(kind in ("1D", "3D"), '"LUT" type must be "1D" or "3D"!')

        is_3D = kind == "3D"

        seek = 2
        metadata = []
        is_metadata = False
        for i, line in enumerate(lines[2:]):
            line = line.strip()
            if line == "BEGIN METADATA":
                is_metadata = True
                continue
            elif line == "END METADATA":
                seek += i
                break

            if is_metadata:
                metadata.append(line)

        title, comments = _parse_metadata_section(metadata)

        seek += 1
        pre_LUT = _parse_domain_section(lines[seek : seek + 9])

        seek += 9
        size, table = _parse_table_section(lines[seek:])

        attest(np.product(size) == len(table), '"LUT" table size is invalid!')

    LUT: Union[LUT3x1D, LUT3D, LUTSequence]
    if (
        is_3D
        and pre_LUT.shape == (6, 2)
        and np.array_equal(
            np.transpose(np.reshape(pre_LUT, (3, 4)))[2:4], unity_range
        )
    ):
        table = table.reshape([size[0], size[1], size[2], 3], order="F")
        LUT = LUT3D(
            domain=np.transpose(np.reshape(pre_LUT, (3, 4)))[0:2],
            name=title,
            comments=comments,
            table=table,
        )

    elif (
        not is_3D
        and pre_LUT.shape == (6, 2)
        and np.array_equal(
            np.transpose(np.reshape(pre_LUT, (3, 4)))[2:4], unity_range
        )
    ):
        LUT = LUT3x1D(
            domain=pre_LUT.reshape(3, 4).transpose()[0:2],
            name=title,
            comments=comments,
            table=table,
        )

    elif is_3D:
        pre_domain = tstack((pre_LUT[0], pre_LUT[2], pre_LUT[4]))
        pre_table = tstack((pre_LUT[1], pre_LUT[3], pre_LUT[5]))
        shaper_name = f"{title} - Shaper"
        cube_name = f"{title} - Cube"
        table = table.reshape([size[0], size[1], size[2], 3], order="F")

        LUT = LUTSequence(
            LUT3x1D(pre_table, shaper_name, pre_domain),
            LUT3D(table, cube_name, comments=comments),
        )

    elif not is_3D:
        pre_domain = tstack((pre_LUT[0], pre_LUT[2], pre_LUT[4]))
        pre_table = tstack((pre_LUT[1], pre_LUT[3], pre_LUT[5]))

        if table.shape == (2, 3):
            table_max = table[1]
            table_min = table[0]
            pre_table *= table_max - table_min
            pre_table += table_min

            LUT = LUT3x1D(pre_table, title, pre_domain, comments=comments)
        else:
            pre_name = f"{title} - PreLUT"
            table_name = f"{title} - Table"

            LUT = LUTSequence(
                LUT3x1D(pre_table, pre_name, pre_domain),
                LUT3x1D(table, table_name, comments=comments),
            )

    return LUT


def write_LUT_Cinespace(
    LUT: Union[LUT3x1D, LUT3D, LUTSequence], path: str, decimals: Integer = 7
) -> Boolean:
    """
    Write given *LUT* to given  *Cinespace* *.csp* *LUT* file.

    Parameters
    ----------
    LUT
        :class:`LUT1D`, :class:`LUT3x1D` or :class:`LUT3D` or
        :class:`LUTSequence` class instance to write at given path.
    path
        *LUT* path.
    decimals
        Formatting decimals.

    Returns
    -------
    :class:`bool`
        Definition success.

    References
    ----------
    :cite:`RisingSunResearch`

    Examples
    --------
    Writing a 3x1D *Cinespace* *.csp* *LUT*:

    >>> from colour.algebra import spow
    >>> domain = np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]])
    >>> LUT = LUT3x1D(
    ...     spow(LUT3x1D.linear_table(16, domain), 1 / 2.2),
    ...     'My LUT',
    ...     domain,
    ...     comments=['A first comment.', 'A second comment.'])
    >>> write_LUT_Cinespace(LUT, 'My_LUT.cube')  # doctest: +SKIP

    Writing a 3D *Cinespace* *.csp* *LUT*:

    >>> domain = np.array([[-0.1, -0.2, -0.4], [1.5, 3.0, 6.0]])
    >>> LUT = LUT3D(
    ...     spow(LUT3D.linear_table(16, domain), 1 / 2.2),
    ...     'My LUT',
    ...     domain,
    ...     comments=['A first comment.', 'A second comment.'])
    >>> write_LUT_Cinespace(LUT, 'My_LUT.cube')  # doctest: +SKIP
    """

    has_3D, has_3x1D = False, False

    if isinstance(LUT, LUTSequence):
        attest(
            len(LUT) == 2
            and isinstance(LUT[0], (LUT1D, LUT3x1D))
            and isinstance(LUT[1], LUT3D),
            '"LUTSequence" must be "1D + 3D" or "3x1D + 3D"!',
        )
        LUT[0] = (
            LUT[0].as_LUT(LUT3x1D) if isinstance(LUT[0], LUT1D) else LUT[0]
        )
        name = f"{LUT[0].name} - {LUT[1].name}"
        has_3x1D = True
        has_3D = True

    elif isinstance(LUT, LUT1D):
        name = LUT.name
        has_3x1D = True
        LUT = LUTSequence(LUT.as_LUT(LUT3x1D), LUT3D())

    elif isinstance(LUT, LUT3x1D):
        name = LUT.name
        has_3x1D = True
        LUT = LUTSequence(LUT, LUT3D())

    elif isinstance(LUT, LUT3D):
        name = LUT.name
        has_3D = True
        LUT = LUTSequence(LUT3x1D(), LUT)

    else:
        raise ValueError("LUT must be 1D, 3x1D, 3D, 1D + 3D or 3x1D + 3D!")

    if has_3x1D:
        attest(
            2 <= LUT[0].size <= 65536,
            "Shaper size must be in domain [2, 65536]!",
        )
    if has_3D:
        attest(
            2 <= LUT[1].size <= 256, "Cube size must be in domain [2, 256]!"
        )

    def _ragged_size(table: ArrayLike) -> List:
        """Return the ragged size of given table."""

        R, G, B = tsplit(table)

        R_len = R.shape[-1] - np.sum(np.isnan(R))
        G_len = G.shape[-1] - np.sum(np.isnan(G))
        B_len = B.shape[-1] - np.sum(np.isnan(B))

        return [R_len, G_len, B_len]

    def _format_array(array: Union[List, Tuple]) -> str:
        """Format given array as a *Cinespace* *.cube* data row."""

        return "{1:0.{0}f} {2:0.{0}f} {3:0.{0}f}".format(decimals, *array)

    def _format_tuple(array: Union[List, Tuple]) -> str:
        """
        Format given array as 2 space separated values to *decimals*
        precision.
        """

        return "{1:0.{0}f} {2:0.{0}f}".format(decimals, *array)

    with open(path, "w") as csp_file:
        csp_file.write("CSPLUTV100\n")

        if has_3D:
            csp_file.write("3D\n\n")
        else:
            csp_file.write("1D\n\n")

        csp_file.write("BEGIN METADATA\n")
        csp_file.write(f"{name}\n")

        if LUT[0].comments:
            for comment in LUT[0].comments:
                csp_file.write(f"{comment}\n")

        if LUT[1].comments:
            for comment in LUT[1].comments:
                csp_file.write(f"{comment}\n")

        csp_file.write("END METADATA\n\n")

        if has_3D:
            if has_3x1D:
                for i in range(3):
                    size = (
                        _ragged_size(LUT[0].domain)[i]
                        if LUT[0].is_domain_explicit()
                        else LUT[0].size
                    )

                    csp_file.write(f"{size}\n")

                    for j in range(size):
                        entry = (
                            LUT[0].domain[j][i]
                            if LUT[0].is_domain_explicit()
                            else (
                                LUT[0].domain[0][i]
                                + j
                                * (LUT[0].domain[1][i] - LUT[0].domain[0][i])
                                / (LUT[0].size - 1)
                            )
                        )

                        csp_file.write("{0:.{1}f} ".format(entry, decimals))

                    csp_file.write("\n")

                    for j in range(size):
                        entry = LUT[0].table[j][i]
                        csp_file.write("{0:.{1}f} ".format(entry, decimals))

                    csp_file.write("\n")
            else:
                for i in range(3):
                    csp_file.write("2\n")
                    domain = _format_tuple(
                        [LUT[1].domain[0][i], LUT[1].domain[1][i]]
                    )
                    csp_file.write(f"{domain}\n")
                    csp_file.write(
                        "{0:.{2}f} {1:.{2}f}\n".format(0, 1, decimals)
                    )

            csp_file.write(
                f"\n{LUT[1].table.shape[0]} "
                f"{LUT[1].table.shape[1]} "
                f"{LUT[1].table.shape[2]}\n"
            )
            table = LUT[1].table.reshape([-1, 3], order="F")

            for row in table:
                csp_file.write(f"{_format_array(row)}\n")
        else:
            for i in range(3):
                csp_file.write("2\n")
                domain = _format_tuple(
                    [LUT[0].domain[0][i], LUT[0].domain[1][i]]
                )
                csp_file.write(f"{domain}\n")
                csp_file.write("0.0 1.0\n")
            csp_file.write(f"\n{LUT[0].size}\n")
            table = LUT[0].table

            for row in table:
                csp_file.write(f"{_format_array(row)}\n")

    return True



"""Defines the unit tests for the :mod:`colour.recovery.smits1999` module."""

import numpy as np
import unittest

from colour.colorimetry import sd_to_XYZ_integration
from colour.recovery import RGB_to_sd_Smits1999
from colour.recovery.smits1999 import XYZ_to_RGB_Smits1999
from colour.utilities import domain_range_scale

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestRGB_to_sd_Smits1999",
]


class TestRGB_to_sd_Smits1999(unittest.TestCase):
    """
    Define :func:`colour.recovery.smits1999.RGB_to_sd_Smits1999`
    definition unit tests methods.
    """

    def test_RGB_to_sd_Smits1999(self):
        """
        Test :func:`colour.recovery.smits1999.RGB_to_sd_Smits1999`
        definition.
        """

        np.testing.assert_almost_equal(
            RGB_to_sd_Smits1999(
                XYZ_to_RGB_Smits1999(
                    np.array([0.21781186, 0.12541048, 0.04697113])
                )
            ).values,
            np.array(
                [
                    0.07691923,
                    0.05870050,
                    0.03943195,
                    0.03024978,
                    0.02750692,
                    0.02808645,
                    0.34298985,
                    0.41185795,
                    0.41185795,
                    0.41180754,
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_sd_Smits1999(
                XYZ_to_RGB_Smits1999(
                    np.array([0.15434689, 0.22960951, 0.09620221])
                )
            ).values,
            np.array(
                [
                    0.06981477,
                    0.06981351,
                    0.07713379,
                    0.25139495,
                    0.30063408,
                    0.28797045,
                    0.11990414,
                    0.08186170,
                    0.08198613,
                    0.08272671,
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_sd_Smits1999(
                XYZ_to_RGB_Smits1999(
                    np.array([0.07683480, 0.06006092, 0.25833845])
                )
            ).values,
            np.array(
                [
                    0.29091152,
                    0.29010285,
                    0.26572455,
                    0.13140471,
                    0.05160646,
                    0.05162034,
                    0.02765638,
                    0.03199188,
                    0.03472939,
                    0.03504156,
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_sd_Smits1999(
                XYZ_to_RGB_Smits1999(np.array([0.0, 1.0, 0.0]))
            ).values,
            np.array(
                [
                    -0.2549796,
                    -0.2848386,
                    -0.1634905,
                    1.5254829,
                    1.9800433,
                    1.8510762,
                    -0.7327702,
                    -1.2758621,
                    -1.2758621,
                    -1.2703551,
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_sd_Smits1999(
                XYZ_to_RGB_Smits1999(np.array([1.0, 1.0, 0.0]))
            ).values,
            np.array(
                [
                    -0.1168428,
                    -0.1396982,
                    -0.0414535,
                    0.581391,
                    0.9563091,
                    0.9562111,
                    1.3366949,
                    1.3742666,
                    1.3853491,
                    1.4027005,
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_sd_Smits1999(
                XYZ_to_RGB_Smits1999(np.array([0.5, 0.0, 1.0]))
            ).values,
            np.array(
                [
                    1.1938776,
                    1.1938776,
                    1.1213867,
                    -0.067889,
                    -0.4668587,
                    -0.4030985,
                    0.703056,
                    0.9407334,
                    0.9437298,
                    0.9383386,
                ]
            ),
            decimal=7,
        )

    def test_domain_range_scale_RGB_to_sd_Smits1999(self):
        """
        Test :func:`colour.recovery.smits1999.RGB_to_sd_Smits1999`
        definition domain and range scale support.
        """

        XYZ_i = np.array([0.20654008, 0.12197225, 0.05136952])
        RGB_i = XYZ_to_RGB_Smits1999(XYZ_i)
        XYZ_o = sd_to_XYZ_integration(RGB_to_sd_Smits1999(RGB_i))

        d_r = (("reference", 1, 1), ("1", 1, 0.01), ("100", 100, 1))
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    sd_to_XYZ_integration(
                        RGB_to_sd_Smits1999(RGB_i * factor_a)
                    ),
                    XYZ_o * factor_b,
                    decimal=7,
                )


if __name__ == "__main__":
    unittest.main()
